#ifndef __GENOTYPEDECODER_H
#define __GENOTYPEDECODER_H

#include <ishtar/settings.h>
#include <an/Geometry.h>
#include "../utils/Graph.h"

/*!	\file GenotypeDecoder.h
	\brief Interfaces of generic genotype decoders
*/

/*!	\defgroup genotypeDecoders Genotype decoders
	The decoders that makes controllers out of genomes
*/

namespace Teem
{
	class Controller;
	class Genome;

	/**
	Interface for genotype decoder. Those classes are responsible for decoding genotype
	into a controller.
	\ingroup genotypeDecoders
	*/
	class GenotypeDecoder
	{
	public:
		//! Constructor, 0 controller input/output
		GenotypeDecoder(const std::string &root) { inputCount = 0; outputCount = 0; this->root = root; }
		//! Virtual destructor, do nothing
		virtual ~GenotypeDecoder() {}

		//! Setup controller IO related parameters.
		virtual void controllerIOSetup(unsigned inCount, unsigned outCount) { inputCount = inCount; outputCount = outCount; }

		//! Create a random genome that is compatible with this decoder.
		virtual Genome* createGenome() = 0;

		//! Decode a genome to a controller.
		virtual Controller* decode(Genome *genome) = 0;

	protected:

		unsigned inputCount; //!< Number of controller inputs
		unsigned outputCount; //!< Number of controller ouputs
		std::string root; //!< Ishtar root
	};

	/**
	Interface for MinSim genotype decoder. Those classes are responsible for decoding genotype.
	\ingroup genotypeDecoders
	*/
	class DummyGenotypeDecoder : public GenotypeDecoder
	{
	public:
		//! Constructor.
		DummyGenotypeDecoder(const std::string &root);
		//! Destructor.
		virtual ~DummyGenotypeDecoder() { }
		//! Don't have a controller in this case.
		virtual Controller* decode(Genome *genome) { return 0; }

		//! Create a random genome that is compatible with this decoder.
		virtual Genome* createGenome();
		//! Set the requested genome length
		void setGenomeLength(size_t len) { genomeLength = len; }

	protected:
		size_t genomeLength; //!< Requested genome length
	};

	/**
	Interface for genotype decoder that create graph-based controllers. It provide
	a hook to allow classes to access the graph before controller creation. This
	allows, for example, experiment class to modify/expand/etc. the graph before the
	actual neural network is created. Sub-classes should call \c callFilter before 
	creating the controller.
	\ingroup genotypeDecoders
	*/
	class GraphGenotypeDecoder : public GenotypeDecoder
	{
	public:
		//! Constructor, graphFilter = NULL
		GraphGenotypeDecoder(const std::string &root) : GenotypeDecoder(root) { graphFilter = NULL; }
		//! Virtual destructor, do nothing
		virtual ~GraphGenotypeDecoder() {}
		
		//! Set a filter for the graph.
		void setGraphFilter(Graph::Filter *filter) { graphFilter = filter; }
	
	protected:
		Graph::Filter* graphFilter; //!< Pointer to a graph filter that can be called before creating the controller
		
		//! Call the graph filter. Subclasses should call this function before
		//! creating the controller from the graph.
		void callFilter(Graph* graph) { if (graphFilter) graphFilter->filter(graph); }
	};
}

#endif
