#ifndef __MATRIXCONTROLLER_H
#define __MATRIXCONTROLLER_H

#include "Controller.h"
#include "GenotypeDecoder.h"
#include "../utils/Matrix.h"
#include <assert.h>
#include <valarray>

/*!	\file MatrixController.h
	\brief Interfaces of the matrix controller
*/

namespace Teem
{
	//! Simple controller described by a matrix. The output value
	//! is a linear combination of the inputs, with coefficient contained
	//! in a matrix.
	//! \ingroup controllers
	class MatrixController : public Controller
	{
	public:
		Matrix<double> matrix; //!< Coefficient of the controller.
		std::valarray<double> bias; //!< Bias vector that will be added after the matrix product
	
	public:
		//! Constructor, create a matrix controller with inputCount inputs and outputCount outputs
		MatrixController(unsigned inputCount, unsigned outputCount);
		//! Destructor
		virtual ~MatrixController() {}
		
		virtual void setInput(unsigned index, double val)
		{
			assert(index < input.size());
			input[index] = val;
		}
		
		virtual double getOutput(unsigned index)
		{
			assert(index < output.size());
			return output[index];
		}
		
		virtual void step();
	
	protected:
	
		//! Storage for input.
		std::valarray<double> input;
		
		//! Storage for output.
		std::valarray<double> output;
	};
	
	class Genome;
	
	//! Genotype decoder for matrix controller.
	//! \ingroup genotypeDecoders
	class MatrixGenotypeDecoder : public GenotypeDecoder
	{
	protected:
		Ishtar::Variable<double> coefRange; //!< Range of coefficient's values, which will be within [-coefRange .. coefRange]
		Ishtar::Variable<unsigned> evolveBias; //!< Non-zero if bias have to be evolved as well // FIXME: bool
		
	public:
		//! Constructor, by default create set coefficient range to 1 and enable evolution of bias also
		MatrixGenotypeDecoder(const std::string &root) :
			GenotypeDecoder(root),
			coefRange(root + ".coefRange", 1.0),
			evolveBias(root + ".evolveBias", 1) { }
		//! Destructor
		virtual ~MatrixGenotypeDecoder() { }
		
		virtual Genome* createGenome();
		virtual Controller* decode(Genome* genome);
	};
}

#endif
