#include <an/Random.h>
#include <algorithm>

#include "FeedForwardNeuralNetwork.h"
#include "../../ga/Genome.h"
#include "../../ga/ValueGenome.h"
#include "../../main/VerboseLevel.h"

/*!	\file FeedForwardNeuralNetwork.cpp
	\brief Implementation of feed-forward only neural networks
*/

namespace Teem
{
	FeedForwardNeuralNetwork::FeedForwardNeuralNetwork(size_t inputCount, size_t outputCount, const std::string &root) :
		hiddenLayerCount(root + ".hiddenLayerCount", 2),
		biasValue(root + ".biasValue", -1),
		activationFunction(root + ".activationFunction", "tanh"),
		activationFunctionParameter(root + ".activationFunctionParameter", 1.0)
	{
		assert(inputCount > 0);
		
		this->inputCount = inputCount;
		this->outputCount = outputCount;
		
		// setup variable for layers
		for(unsigned i = 0; i < hiddenLayerCount; i++)
		{
			std::ostringstream oss;
			oss << root << ".hiddenLayer" << i << "Size";
			layerSizes.push_back(Ishtar::Variable<unsigned>(oss.str(), 2));
		}
		layerSizes.push_back(outputCount);
		layerCount = hiddenLayerCount + 1;
		
		// init state variables
		for(unsigned i = 0; i < layerCount; i++)
		{
			unsigned layerInCount = (i == 0 ? inputCount : layerSizes[i-1]);
			unsigned layerOutCount = layerSizes[i];
			weights.push_back(Matrix<double>(layerInCount, layerOutCount));
			biasWeights.push_back(std::valarray<double>(layerOutCount));
			activations.push_back(std::valarray<double>(layerOutCount));
			outputs.push_back(std::valarray<double>(layerOutCount));
		}
		
		input.resize(inputCount);
		
		std::string s = activationFunction;
		if(s == "tanh")
			forwardActFunc = &TanhForwardActivationFunction;
		else
			assert(false);
	}
	
	void FeedForwardNeuralNetwork::setInput(unsigned index, double in)
	{
		assert(index < input.size());
		
		input[index] = in;
	}
	
	double FeedForwardNeuralNetwork::getOutput(unsigned index)
	{
		assert(index < outputs[outputs.size()-1].size());
		
		return outputs[outputs.size()-1][index];
	}
	
	void FeedForwardNeuralNetwork::step()
	{
		// for each layer...
		for(unsigned i = 0; i < layerCount; i++)
		{
			// output from previous layer
			std::valarray<double> &layerInput = (i == 0 ? input : outputs[i-1]);
			
			// compute activation for this layer
			activations[i] = weights[i] * layerInput;
			activations[i] += biasWeights[i] * ((double) biasValue);
			
			// compute output for this layer
			std::for_each(&activations[i][0], &activations[i][activations[i].size()], ActivationFunctor(forwardActFunc, outputs[i], activationFunctionParameter));
		}
	}
	
	void FeedForwardNeuralNetwork::randomize(double from, double to)
	{
		for(size_t i = 0; i < weights.size(); i++)
		{
			std::generate(&weights[i].flat()[0], &weights[i].flat()[weights[i].flat().size()], An::UniformRand(from, to));
			std::generate(&biasWeights[i][0], &biasWeights[i][biasWeights[i].size()], An::UniformRand(from, to));
		}
	}
	
	// ======================================================
	// ======================================================
	
	RecurrentFeedForwardNeuralNetwork::RecurrentFeedForwardNeuralNetwork(size_t inputCount, size_t outputCount, const std::string &root) :
		FeedForwardNeuralNetwork(inputCount, outputCount, root),
		outputLayerRecursive(root + ".outputLayerRecursive", false)
	{
		// hidden layers
		for(unsigned i = 0; i < hiddenLayerCount; i++)
		{
			std::ostringstream oss; oss << root << ".hiddenLayer" << i << "Recursive";
			Ishtar::Variable<bool> recurs(oss.str(), false);
			layerRecursive.push_back(recurs);
			
			if(recurs)
				recursiveWeights.push_back(Matrix<double>(layerSizes[i], layerSizes[i]));
			else
				recursiveWeights.push_back(Matrix<double>(0, 0));
		}
		
		// output layer
		layerRecursive.push_back(outputLayerRecursive);
		if(outputLayerRecursive)
			recursiveWeights.push_back(Matrix<double>(outputCount, outputCount));
		else
			recursiveWeights.push_back(Matrix<double>(0, 0));
	}
	
	void RecurrentFeedForwardNeuralNetwork::step()
	{
		// for each layer...
		for(unsigned i = 0; i < layerCount; i++)
		{
			// output from previous layer
			std::valarray<double> &layerInput = (i == 0 ? input : outputs[i-1]);
			
			// compute activation for this layer
			activations[i] = weights[i] * layerInput;
			activations[i] += biasWeights[i] * ((double) biasValue);
			if(layerRecursive[i])
				activations[i] += recursiveWeights[i] * outputs[i];
			
			// compute output for this layer
			std::for_each(&activations[i][0], &activations[i][activations[i].size()], ActivationFunctor(forwardActFunc, outputs[i], activationFunctionParameter));
		}
	}
	
	void RecurrentFeedForwardNeuralNetwork::randomize(double from, double to)
	{
		// call inherited
		FeedForwardNeuralNetwork::randomize(from, to);
		
		// randomize recursive weights
		for(size_t i = 0; i < hiddenLayerCount; i++)
		{
			if(layerRecursive[i])
				std::generate(&recursiveWeights[i].flat()[0], &recursiveWeights[i].flat()[recursiveWeights[i].flat().size()], An::UniformRand(from, to));
		}
	}
	
	// ======================================================
	// ======================================================
	
	BackPropFeedForwardNeuralNetwork::BackPropFeedForwardNeuralNetwork(size_t inputCount, size_t outputCount, const std::string &root) :
		FeedForwardNeuralNetwork(inputCount, outputCount, root),
		learningRate(root + ".learningRate", 0.1)
	{
		// init state variables
		for(unsigned i = 0; i < layerCount; i++)
			deltas.push_back(std::valarray<double>( layerSizes[i]));
		
		error.resize(outputCount);
		
		std::string s = activationFunction;
		if(s == "tanh")
			backwardActFunc = &TanhBackwardActivationFunction;
		else
			assert(false);
	}
	
	void BackPropFeedForwardNeuralNetwork::stepBackward()
	{	
		// Compute deltas for output layer
		size_t idx = activations.size()-1;
		std::for_each(&activations[idx][0], &activations[idx][activations[idx].size()], ActivationFunctor(backwardActFunc, deltas[idx], activationFunctionParameter));
		//deltas[idx] += 0.1;	// suggested by Fahlman (1989), as mentionned in Herz, Krogh & Palmer, pg. 122
		deltas[idx] *= error;
		
		// Compute deltas for every other layers
		for(int i = idx-1; i >= 0; i--)
		{
			std::for_each(&activations[i][0], &activations[i][activations[i].size()], ActivationFunctor(backwardActFunc, deltas[i], activationFunctionParameter));
			deltas[i] *= deltas[i+1] * weights[i+1];
		}
		
		// Update all weights
		for(unsigned i = 0; i < weights.size(); i++)
		{
			Matrix<double> &W = weights[i];
			size_t fromCount = W.columnNum();
			size_t toCount = W.rowNum();
			std::valarray<double> &fromLayer = (i == 0 ? input : outputs[i-1]);
			
			// update weights coming from previous layer
			for(size_t to = 0; to < toCount; to++)
				for(size_t from = 0; from < fromCount; from++)
				{
					W(from, to) += learningRate * deltas[i][to] * fromLayer[from];
				}
			
			// update weights coming from bias.
			biasWeights[i] += learningRate * biasValue * deltas[i];
		}
	}
	
	// ======================================================
	// ======================================================
	
	FeedForwardNeuralNetworkGenotypeDecoder::FeedForwardNeuralNetworkGenotypeDecoder(const std::string &root) :
		GenotypeDecoder(root),
		hiddenLayerCount(root + ".hiddenLayerCount", 2),
		useBackProp(root + ".useBackProp", false),
		useRecursion(root + ".useRecursion", false),
		weightRange(root + ".weightRange", 2.0)
	{
		if(useBackProp && useRecursion)
		{
			CERR("!!! FeedForwardNeuralNetworkGenotypeDecoder::ctor : Back-propagation and recursive FF NN are NOT compatible. Not using backpropagation");
			useBackProp = false;
		}
		
		for(unsigned i = 0; i < hiddenLayerCount; i++)
		{
			std::ostringstream oss;
			oss << root << ".hiddenLayer" << i << "Size";
			hiddenLayerSizes.push_back(Ishtar::Variable<unsigned>(oss.str(), 2));
			
			if(useRecursion)
			{
				std::ostringstream oss2; oss2 << root << ".hiddenLayer" << i << "Recursive";
				layerRecursive.push_back(Ishtar::Variable<bool>(oss2.str(), false));
			}
			else
				layerRecursive.push_back(false);
		}
		
		if(useRecursion)
			layerRecursive.push_back(Ishtar::Variable<bool>(root + ".outputLayerRecursive", false));
		else
			layerRecursive.push_back(false);
	}
	
	Genome* FeedForwardNeuralNetworkGenotypeDecoder::createGenome()
	{
		unsigned geneCount = 0;

		// count genes
		for(size_t layer = 0; layer < hiddenLayerSizes.size()+1; layer++)
		{
			size_t fromCount = layer == 0 ? inputCount : hiddenLayerSizes[layer-1];
			size_t toCount = layer == hiddenLayerCount ? outputCount : hiddenLayerSizes[layer];
			
			geneCount += (fromCount + 1) * toCount;
			
			if(layerRecursive[layer])
				geneCount += toCount * toCount;
		}
		
		// Create genome
		Ishtar::Variable<unsigned> defaultPrecision("teem.ga.defaultPrecision", 8);
		Genome* genome = Genome::create(Genome::BIT_GENOME, geneCount * defaultPrecision);
		genome->randomize();
		return genome;
	}
		
	Controller* FeedForwardNeuralNetworkGenotypeDecoder::decode(Genome *genome)
	{
		ValueGenome *vg = dynamic_cast<ValueGenome*>(genome);
		if(vg == NULL) return NULL;
		
		// Create appropriate neural network object.
		FeedForwardNeuralNetwork* nn;
		if(useBackProp)
			nn = new BackPropFeedForwardNeuralNetwork(inputCount, outputCount, root);
		else if(useRecursion)
			nn = new RecurrentFeedForwardNeuralNetwork(inputCount, outputCount, root);
		else
			nn = new FeedForwardNeuralNetwork(inputCount, outputCount, root);
		
		// Set the weight appropriately.
		for(size_t layer = 0; layer < hiddenLayerCount+1; layer++)
		{
			size_t fromCount = layer == 0 ? inputCount : hiddenLayerSizes[layer-1];
			size_t toCount = layer == hiddenLayerCount ? outputCount : hiddenLayerSizes[layer];
			for(size_t to = 0; to < toCount; to++)
			{
				for(size_t from = 0; from < fromCount; from++)	
					nn->setWeight(layer, from, to, vg->getDouble(-weightRange, weightRange));
				nn->setBiasWeight(layer, to, vg->getDouble(-weightRange, weightRange));
			}
			
			if(layerRecursive[layer])
			{
				for(size_t from = 0; from < toCount; from++)
					for(size_t to = 0; to < toCount; to++)
						static_cast<RecurrentFeedForwardNeuralNetwork*>(nn)->setRecursiveWeight(layer, from, to, vg->getDouble(-weightRange, weightRange));
			}
		}
		
		vg->resetReading();
		
		return nn;
	}
}
