#ifndef _FEEDFORWARDNN_H
#define _FEEDFORWARDNN_H

#include "../Controller.h"
#include "../../utils/Matrix.h"
#include "../GenotypeDecoder.h"
#include <valarray>
#include <vector>
#include <ishtar/settings.h>

/*!	\file FeedForwardNeuralNetwork.h
	\brief Interface of feed-forward only neural networks
*/

namespace Teem
{
	//! Class that implement a simple multilayer feedforward neural network
	//! without using the abstract NN architecture.
	//! \ingroup controllers
	class FeedForwardNeuralNetwork : public Controller
	{
	public:
		//! Constructor, create a neural network with inputCount inputs and outputCount outputs
		FeedForwardNeuralNetwork(size_t inputCount, size_t outputCount, const std::string &root);
		//! Destructor
		virtual ~FeedForwardNeuralNetwork() { }
		
		// Controller overrides.
		//! Set the input values to the neural network.
		virtual void setInput(unsigned index, double val);
		//! Return the input value.
		virtual double getInput(unsigned index) const { assert(index < inputCount); return input[index]; }
		//! Returns the number of input.
		virtual unsigned getInputCount() const { return inputCount; }
		//! Read the output values.
		virtual double getOutput(unsigned index);
		//! Propagate the input values to the output through all the layers.
		virtual void step();
		
		//! Return the number of layers.
		size_t layerNum() { return layerCount; }
		//! Return the size of the nth layer. Layer 0 is the fist layer after input.
		size_t layerSize(size_t layer) { return layerSizes[layer]; }
		//! Return the number of input.
		size_t inputNum() { return inputCount; }
		//! Return the number of output.
		size_t outputNum() { return outputCount; }
		
		//! Set the weight of a particular synapse to a particular layer. Using layer 0,
		//! synpases from input to layer 0 are set.
		void setWeight(size_t toLayer, size_t from, size_t to, double w) { weights[toLayer](from, to) = w; }
		//! Get the value for a particular weight.
		double getWeight(size_t toLayer, size_t from, size_t to) const { return weights[toLayer](from, to); }
		
		//! Set weight from bias to neuron.
		void setBiasWeight(size_t toLayer, size_t to, double w) { biasWeights[toLayer][to] = w; }
		//! Get weight from bias to neuron.
		double getBiasWeight(size_t toLayer, size_t to) const { return biasWeights[toLayer][to]; }
		
		//! Put random weights with uniform distribution.
		virtual void randomize(double from, double to);
	
	protected:
		
		//! Activation function y = g(x)
		//! @param x activation
		//! @param b activation function parameter (slope)
		//! @return output
		typedef double (*ActivationFunction)(double x, double b);
	
		class ActivationFunctor;
		
		size_t inputCount; //!< number of input
		size_t outputCount; //!< number of output
		size_t layerCount; //!< number of layer
		Ishtar::Variable<unsigned> hiddenLayerCount;  //!< number of hidden layer
		std::vector<size_t> layerSizes;  //!< size of layers
		Ishtar::Variable<double> biasValue;  //!< value of the bias neuron
		Ishtar::Variable<std::string> activationFunction;  //!< name of the activation function
		Ishtar::Variable<double> activationFunctionParameter;  //!< parameter for the activation function
		
		std::vector<Matrix<double> > weights;  //!< weight matrix for each layer
		std::vector<std::valarray<double> > biasWeights;  //!< weight from the bias to each layer
		std::vector<std::valarray<double> > activations;  //!< activation of each neuron
		std::vector<std::valarray<double> > outputs;  //!< output of each neuron
		std::valarray<double> input;  //!< input vector
		ActivationFunction forwardActFunc;  //!< pointer to activation function
		
		//! Tanh activation function.
		static double TanhForwardActivationFunction(double x, double b)
		{
			return tanh(x * b);
		}
	};
	
	//! Functor to compute activation function on std::valarray<double>
	//! unsing std::for_each.
	class FeedForwardNeuralNetwork::ActivationFunctor
	{
	protected:
		ActivationFunction function;  //!< pointer to activation function
		double b;  //!< activation function parameter
		std::valarray<double>& result;  //!< array to hold result
		size_t pos;  //!< position in the result array
		
	public:
		//! Constructor, use function func and parameter b, and return result in res and 
		ActivationFunctor(ActivationFunction func, std::valarray<double> &res, double b)
			: result(res)
		{
			this->function = func;
			this->pos = 0;
			this->b = b;
		}
		
		//! Do the operation, store value in result[pos] and increment pos
		void operator()(double x)
		{
			result[pos++] = function(x, b);
		}
	};
	
	
	/** Feed forward neural network with optional recurence for each layer. \ingroup controllers */
	class RecurrentFeedForwardNeuralNetwork : public FeedForwardNeuralNetwork
	{
	public:
		//! Constructor, create a recurrent neural network with inputCount inputs and outputCount outputs
		RecurrentFeedForwardNeuralNetwork(size_t inputCount, size_t outputCount, const std::string &root);
		//! Destructor
		virtual ~RecurrentFeedForwardNeuralNetwork() { }
		
		//! Propagate the input values to the output through all the layers.
		virtual void step();
		//! Put random weights with uniform distribution.
		virtual void randomize(double from, double to);
		
		//! Set the recursive weight in a layer.
		void setRecursiveWeight(size_t layer, size_t from, size_t to, double w) { recursiveWeights[layer](from, to) = w; }
		//! Get the value for a particular recursive weight.
		double getRecursiveWeight(size_t layer, size_t from, size_t to) const { return recursiveWeights[layer](from, to); }
		//! Return true if layer is recurrent.
		bool isLayerRecursive(size_t layer) { return layerRecursive[layer]; }
	
	protected:
	
		Ishtar::Variable<bool> outputLayerRecursive; //!< is output layer recursive?
		std::vector<Matrix<double> > recursiveWeights; //!< recusive weights for each layer
		std::vector<bool> layerRecursive; //!< is layer recursive?
	};
	
	
	/**
	Feed forward neural network with online back-propagation. Use that way:
	
\code
for i in inputs
	nn.setInput(i, inputValue);
nn.step();Ishtar::Variable<bool> outputLayerRecursive; //! is output layer recursive?
for i in outputs
	nn.setError(i, errorValue);
nn.stepBackward();
\endcode
	\ingroup controllers
	*/
	class BackPropFeedForwardNeuralNetwork : public FeedForwardNeuralNetwork
	{
	public:
		//! Constructor, create a feed forward neural network with online back-propagation with inputCount inputs and outputCount outputs
		BackPropFeedForwardNeuralNetwork(size_t inputCount, size_t outputCount, const std::string &root);
		//! Destructor
		virtual ~BackPropFeedForwardNeuralNetwork() { }
		
		//! Backpropagate the error on the weight.
		virtual void stepBackward();
		
		//! Set the output error (desired value minus obtained value)
		void setError(size_t index, double val)
		{
			assert(index < error.size());
			error[index] = val;
		}
		
		//! Set the learning rate value.
		void setLearningRate(double rate) { learningRate = rate; }
		//! Get the learning rate value.
		double getLearningRate() { return learningRate; }
	
	protected:
		
		Ishtar::Variable<double> learningRate;  //!< learning rate constant
		std::valarray<double> error;  //!< desired output
		ActivationFunction backwardActFunc;  //!< derivative of the activation function
		std::vector<std::valarray<double> > deltas;  //!< deltas (see backprop algorithm)
	
		//! Tanh backward activation function.
		static double TanhBackwardActivationFunction(double x, double b)
		{
			double t = tanh(x * b);
			return b * (1.0 - t * t);
		}
	};
	
	
	/** Genotype decoder for feedforward NN. Can be used with either the simple
	one or the back-propagation version. In the later case, the initial weight are 
	encoded but subject to modification by back-prop.
	\ingroup genotypeDecoders
	*/
	class FeedForwardNeuralNetworkGenotypeDecoder : public GenotypeDecoder
	{
	public:
		//! Constructor
		FeedForwardNeuralNetworkGenotypeDecoder(const std::string &root);
		//! Destructor
		virtual ~FeedForwardNeuralNetworkGenotypeDecoder() { }
		
		//! Create a random genome that is compatible with this decoder.
		virtual Genome* createGenome();
		
		//! Decode a genome to a controller.
		virtual Controller* decode(Genome *genome);
	
	protected:
	
		Ishtar::Variable<unsigned> hiddenLayerCount;  //!< number of hidden layer
		std::vector<size_t> hiddenLayerSizes;  //!< size of hidden layers
		Ishtar::Variable<bool> useBackProp; //!< use simple FF or backprop?
		Ishtar::Variable<bool> useRecursion; //!< use simple FF or recursive FF?
		Ishtar::Variable<double> weightRange; //!< weight range to use for decoding
		std::vector<bool> layerRecursive; //!< is the layer recursive?
 	};
}

#endif
