
#include "GraphNeuralNetwork.h"
#include "Neuron.h"
#include "../../main/Registry.h"
#include "../../main/VerboseLevel.h"
#include "../../ga/ValueGenome.h"

/*!	\file GraphNeuralNetwork.cpp
	\brief Implementation of neural networks based on graphs
*/

namespace Teem
{
	GraphNeuralNetwork::GraphNeuralNetwork(const Graph &graph, const std::string &root)
	{
		// Build neuron.
		neurons.resize(graph.nodes.size());
		for(size_t i = 0; i < graph.nodes.size(); i++)
		{
			Neuron* neuron = neuronsLibrary->createItem(graph.nodes[i].type);
			neuron->setParams(graph.nodes[i].params);
			neurons[i] = neuron;
		}
		
		// Build synapses.
		synapses.resize(graph.links.size());
		for(size_t i = 0; i < graph.links.size(); i++)
		{
			Synapse* synapse = synapsesLibrary->createItem(graph.links[i].type);
			synapse->init(neurons[graph.links[i].from], neurons[graph.links[i].to]);
			synapse->setParams(graph.links[i].params);
			synapses[i] = synapse;
		}
		
		// Build vector map.
		for(std::map<std::string, std::vector<Graph::Node> >::const_iterator i = graph.nodeLists.begin();
		i != graph.nodeLists.end(); i++)
		{
			neuronLists[i->first].resize(i->second.size());
			for(size_t j = 0; j < i->second.size(); j++)
			{
				neuronLists[i->first][j] = neurons[i->second[j]];
				neurons[i->second[j]]->tags.insert(i->first);
			}
		}
		for(std::map<std::string, std::vector<Graph::Link> >::const_iterator i = graph.linkLists.begin(); i != graph.linkLists.end(); i++)
		{
			synapseLists[i->first].resize(i->second.size());
			for(size_t j = 0; j < i->second.size(); j++)
			{
				synapseLists[i->first][j] = synapses[i->second[j]];
				synapses[i->second[j]]->tags.insert(i->first);
			}
		}
		
		// Call finalize for all the neurons and synapses.
		for(size_t i = 0; i < neurons.size(); i++)
			neurons[i]->finalize();
		for(size_t i = 0; i < synapses.size(); i++)
			synapses[i]->finalize();
		
		// Cache input and output neurons for performances.
		outputNeurons = &neuronLists["output"];
		inputNeurons.resize(neuronLists["input"].size());
		for(size_t i = 0; i < inputNeurons.size(); i++)
		{
			inputNeurons[i] = dynamic_cast<InputNeuron*>(neuronLists["input"][i]);
			if(inputNeurons[i] == NULL)
				COUT(1, "GraphNeuralNetwork::GraphNeuralNetwork :  WARNING : Input neuron " << i << " is not an InputNeuron." << std::endl);
		}
	}
	
	GraphNeuralNetwork::~GraphNeuralNetwork()
	{
		for (size_t i = 0; i < synapses.size(); i++)
			delete synapses[i];
		for (size_t i = 0; i < neurons.size(); i++)
			delete neurons[i];
	}
	
	void GraphNeuralNetwork::setInput(unsigned index, double val)
	{
		assert(index < inputNeurons.size());
		
		InputNeuron *input = inputNeurons[index];
		if (input)
			input->setInputValue(val);
	}
	
	double GraphNeuralNetwork::getOutput(unsigned index)
	{
		assert(index < outputNeurons->size());
		
		return (*outputNeurons)[index]->getCurrent();
	}
	
	void GraphNeuralNetwork::step()
	{
		// First we update all the synaps
		for (size_t i=0; i<synapses.size(); i++)
			synapses[i]->update();
		// Then all the neurons
		for (size_t i=0; i<neurons.size(); i++)
			neurons[i]->update();
		// Then we commit the result to the output
		for (size_t i=0; i<neurons.size(); i++)
			neurons[i]->commit();
	}
	
	Controller* GraphFeedForwardGenotypeDecoder::decode(Genome *genome)
	{
		ValueGenome *vg = dynamic_cast<ValueGenome*>(genome);
		if (vg)
		{
			Graph graph;
			Graph::Node node, bias;
			Graph::Link link;
			
			// create bias
			bias = graph.createNode("CurrentInput", TagSet("bias"));
			graph.addNodeParam(bias, "value", -1);
			
			// Create inputs
			for (size_t i = 0; i < inputCount; i++)
				graph.createNode("CurrentInput", TagSet("input"));
			
			// Create hidden if any
			for (size_t i = 0; i < hiddenCount; i++)
			{
				node = graph.createNode("Sigmoid", TagSet("hidden"));
				
				// Create corresponding synapses.
				for (size_t j = 0; j < inputCount; j++)
				{
					link = graph.createLink(graph.getNode("input", j), node, "Simple", TagSet("inputToHidden"));
					graph.addLinkParam(link, "weight", vg->getDouble(-weightRange, weightRange));
				}
				link = graph.createLink(bias, node, "Simple", TagSet("biasToHidden"));
				graph.addLinkParam(link, "weight", vg->getDouble(-weightRange, weightRange));
			}
			
			// Create outputs
			for (size_t i = 0; i < outputCount; i++)
			{
				node = graph.createNode("Sigmoid", TagSet("output"));
				
				std::string fromTagList;
				std::string tag;
				if (hiddenCount == 0)
				{
					fromTagList = "input";
					tag = "inputToOut";
				}
				else
				{
					fromTagList = "hidden";
					tag = "hiddenToOut";
				}
				
				for (size_t j = 0; j < graph.getNodeCount(fromTagList); j++)
				{
					link = graph.createLink(graph.getNode(fromTagList, j), node, "Simple", TagSet(tag));
					graph.addLinkParam(link, "weight", vg->getDouble(-weightRange, weightRange));
				}
				
				link = graph.createLink(bias, node, "Simple", TagSet(tag));
				graph.addLinkParam(link, "weight", vg->getDouble(-weightRange, weightRange));
			}
			
			vg->resetReading();
			
			// Create controller.
			callFilter(&graph);
			return new GraphNeuralNetwork(graph, root);
		}
		else
			return NULL;
	}

	Genome* GraphFeedForwardGenotypeDecoder::createGenome(void)
	{
		// Compute gene count.
		unsigned geneCount;
		if (hiddenCount == 0)
			geneCount = (inputCount + 1) * outputCount;
		else
			geneCount = (inputCount + 1) * hiddenCount + (hiddenCount + 1) * outputCount;
		
		// Create genome
		Ishtar::Variable<unsigned> defaultPrecision("teem.ga.defaultPrecision", 8);
		Genome* genome = Genome::create(Genome::BIT_GENOME, geneCount * defaultPrecision);
		genome->randomize();
		return genome;
	}

	Controller* GraphFullyConnectedGenotypeDecoder::decode(Genome *genome)
	{
		ValueGenome *vg = dynamic_cast<ValueGenome*>(genome);
		if (vg)
		{
			Graph graph;
			Graph::Node node, bias;
			Graph::Link link;
			
			// create bias
			bias = graph.createNode("CurrentInput", TagSet("bias"));
			graph.addNodeParam(bias, "value", -1);
			
			// Create inputs
			for (size_t i = 0; i < inputCount; i++)
				graph.createNode("CurrentInput", TagSet("input"));
			
			// Create hidden if any
			for (size_t i = 0; i < hiddenCount; i++)
				graph.createNode("Sigmoid", TagSet("hidden"));

			// Create outputs
			for (size_t i = 0; i < outputCount; i++)
				graph.createNode("Sigmoid", TagSet("output"));
			
			// Create hidden synapses
			for (size_t i = 0; i < hiddenCount; i++)
			{
				node = graph.getNode("hidden", i);
				
				// from hidden
				for (size_t j = 0; j < hiddenCount; j++)
				{
					link = graph.createLink(graph.getNode("hidden", j), node, "Simple", TagSet("hiddenToHidden"));
					graph.addLinkParam(link, "weight", vg->getDouble(-weightRange, weightRange));
				}
				
				// from output
				for (size_t j = 0; j < outputCount; j++)
				{
					link = graph.createLink(graph.getNode("output", j), node, "Simple", TagSet("outputToHidden"));
					graph.addLinkParam(link, "weight", vg->getDouble(-weightRange, weightRange));
				}
				
				// from input
				for (size_t j = 0; j < inputCount; j++)
				{
					link = graph.createLink(graph.getNode("input", j), node, "Simple", TagSet("inputToHidden"));
					graph.addLinkParam(link, "weight", vg->getDouble(-weightRange, weightRange));
				}
				
				// from bias
				link = graph.createLink(bias, node, "Simple", TagSet("biasToHidden"));
				graph.addLinkParam(link, "weight", vg->getDouble(-weightRange, weightRange));
			}
			
			// Create outputs synapses
			for (size_t i = 0; i < outputCount; i++)
			{
				node = graph.getNode("output", i);
				
				// from hidden
				for (size_t j = 0; j < hiddenCount; j++)
				{
					link = graph.createLink(graph.getNode("hidden", j), node, "Simple", TagSet("hiddenToOutput"));
					graph.addLinkParam(link, "weight", vg->getDouble(-weightRange, weightRange));
				}
				
				// from output
				for (size_t j = 0; j < outputCount; j++)
				{
					link = graph.createLink(graph.getNode("output", j), node, "Simple", TagSet("outputToOutput"));
					graph.addLinkParam(link, "weight", vg->getDouble(-weightRange, weightRange));
				}
				
				// from input
				for (size_t j = 0; j < inputCount; j++)
				{
					link = graph.createLink(graph.getNode("input", j), node, "Simple", TagSet("inputToOutput"));
					graph.addLinkParam(link, "weight", vg->getDouble(-weightRange, weightRange));
				}
				
				// from bias
				link = graph.createLink(bias, node, "Simple", TagSet("biasToOutput"));
				graph.addLinkParam(link, "weight", vg->getDouble(-weightRange, weightRange));
			}
			
			vg->resetReading();
			
			// Create controller.
			callFilter(&graph);
			return new GraphNeuralNetwork(graph, root);
		}
		else
			return NULL;
	}

	Genome* GraphFullyConnectedGenotypeDecoder::createGenome(void)
	{
		// Compute gene count.
		unsigned geneCount = (outputCount + hiddenCount) * (outputCount + hiddenCount + inputCount + 1);
		
		// Create genome
		Ishtar::Variable<unsigned> defaultPrecision("teem.ga.defaultPrecision", 8);
		Genome* genome = Genome::create(Genome::BIT_GENOME, geneCount * defaultPrecision);
		genome->randomize();
		return genome;
	}
}
