#ifndef __SRMNEURON_H
#define __SRMNEURON_H

#include "Neuron.h"
#include <an/Numeric.h>
#include "../../utils/SpikeDelayLine.h"

#include <queue>
#include <cmath>

/*!	\file SRMNeuron.h
	\brief Interfaces of SRM (Spike Response Model) neurons and synapses
*/

namespace Teem
{
	/**
	This class implement a SRM neuron. It will use the (continuous) response from
	the incoming synapses to update its membrane potential. The noise is implemented
	on the threshold, as a gaussian noise of dev. equal to treshold * thresNoise.
	
	The parameters handled by this class are:
		- \c treshold
		- \c refracQuantum
		- \c tau
		- \c tresholdNoise
	\ingroup neurons
	*/
	class SRMNeuron : public Neuron
	{
	public:
		SRMNeuron();
		virtual ~SRMNeuron() { }
	
		// Neuron overrides
		virtual void setParams(const std::string &param, double value);
		virtual void update(double dt);
		virtual void commit();
		
		//! Function to allow external manipulators (typically experiment) to
		//! override spike emission.
		void overrideSpike(bool spike) { lastSpike = spike; }
	
	public:
		//! membrane potential computed during the las update
		double membrane;
		//! threshold of the neuron
		double threshold;
		
	protected:
		//! state of the refractory part of the model
		An::ExpDecay refractory;
		//! true if last update emitted a spike
		bool isSpiking;
		//! noise factor on the threshold (see class description)
		double thresNoise;
		//! refractory period amplitude
		double refracQuantum;
	};
	
	/**
	Base class for synapses that delay spike before using them. Sub-class
	should call Synapse::update(double) and use incomingSpikes in their own
	update() implementation.
	
	The parameters handled by this class are:
		- \c delay
	\ingroup neurons
	*/
	class DelaySynapse : public Synapse
	{
	public:
		DelaySynapse() { setDelay(2.0); incomingSpikes = 0; }
		virtual ~DelaySynapse() { }
		
		virtual void setParams(const std::string &param, double value);
		virtual void update(double dt = 1.0);
		
		//! Set the synaptic delay.
		virtual void setDelay(double theDelay) { delayLine.setDelay(theDelay); }
	
	protected:
		//! Delay line for the spike.
		SpikeDelayLine delayLine;
		//! Number of incoming spikes during the update.
		unsigned incomingSpikes;
	};
	
	/**
	This class implement an SRM synapse. It is designed to work
	with SRMNeuron. The spike are _not_ forwared, but instead
	converted into the response generated by the SRM kernel.
	The equation of the kernel is:
	\f[\epsilon(s) = 4 \exp(-\frac{s \log(2)}{\tau}) * \exp(-\frac{2 s log(2)}{\tau})\f]
	
		
	The parameters handled by this class are:
		- \c tau
	\ingroup neurons
	*/
	class SRMSynapse :  public DelaySynapse
	{
	public:
		SRMSynapse();
		virtual ~SRMSynapse() {}
	
		// Synapse overrides
		virtual void setParams(const std::string &param, double value);
		virtual bool getSpike() const {return false;}
		virtual double getCurrent() const {return response;}
		
		virtual void update(double dt = 1.0);
		
		// Accessors.
		//! Set the time-constant for the response kernel.
		virtual void setTau(double theTau)
		{
			spikeResponse1.setTau(theTau / std::log(2.0));
			spikeResponse2.setTau(theTau / (2.0 * std::log(2.0)));
		}
	
	protected:
		
		//! First state variable for the synaptic response.
		An::ExpDecay spikeResponse1;
		//! Second state variable for the synaptic response.
		An::ExpDecay spikeResponse2;
		//! Response of the synapse.
		double response;
		
	};
	
	/**
	This class implements a SRM synapse with STDP learning. The STDP kernel
	takes the form of two exponentials.
	
	The parameters handled by this class are:
		- \c potFactor
		- \c deprFactor
		- \c potTau
		- \c deprTau
		- \c learningDelay
	\ingroup neurons
	*/
	class SRMPlasticSynapse : public SRMSynapse
	{
	public:
		SRMPlasticSynapse();
		virtual ~SRMPlasticSynapse() {}
		
		virtual void setParams(const std::string &param, double value);
		
		virtual void update(double dt = 1.0);
		
		static double learningRate;  //!< learning rate global parameter
		static double stdpControl; //!< stdp control global parameter
		
	protected:
	
		//! Scale of potentiation part.
		double potFactor;
		//! Scale of depression part.
		double deprFactor;
		//! State variable for potentiation.
		An::ExpDecay potState;
		//! State variable for depression.
		An::ExpDecay deprState;
		//! Delay line for pre neuron.
		SpikeDelayLine preDelayLine;
		//! Delay line for post neuron.
		SpikeDelayLine postDelayLine;
		
		
		//! Preform learning on synapse.
		void preformLearning(double dt);
	};
}

#endif
