/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "GenerationalEvolution.h"
#include "../controller/GenotypeDecoder.h"
#include "../experiment/Experiment.h"
#include "../main/Registry.h"
#include "../ga/Genome.h"
#include "../ga/ValueGenome.h"
#include <ishtar/network.h>
#include <ishtar/filestream.h>
#include <assert.h>
#include <algorithm>
#include <set>
#include <iostream>
#include <fstream>
#include <sstream>
#include <ostream>
#include <iomanip>
#include <limits>
#include <dirent.h>

/*!	\file GenerationalEvolution.cpp
	\brief Implementation of algorithms for generations based evolution based
*/

namespace Teem
{
	// GenerationalIndividual
	
	GenerationalIndividual::GenerationalIndividual()
	{
		genome = NULL;
		fitness = 0;
		trialCount = 0;
	}
	
	void GenerationalIndividual::reset(void)
	{
		fitness = 0;
		trialCount = 0;
	}
	
	void GenerationalIndividual::normalize(void)
	{
		if (trialCount)
			fitness /= trialCount;
	}
	
	void GenerationalIndividual::load(Ishtar::InputStream *stream)
	{
		fitness = stream->readDouble ();
		trialCount = 0;
		if (genome)
			delete genome;
		genome = Genome::load(stream);
	}
	
	void GenerationalIndividual::save(Ishtar::OutputStream *stream)
	{
		stream->writeDouble(fitness);
		genome->save(stream);
	}
	
	
	// GenerationalEvolutionRun
	
	GenerationalEvolutionRun::GenerationalEvolutionRun() :
		crossoverProbability("teem.ga.crossoverProbability", 0.2),
		elitismRatio("teem.ga.elitismRatio", 0.1),
		reproductionRatio("teem.ga.reproductionRatio", 0.3),
		populationCount("teem.ga.populationCount", 80),
		generationCount("teem.ga.generationCount", 400),
		generation(0)
	{
		// integrity constraints
		assert(crossoverProbability >= 0);
		assert(crossoverProbability <= 1);
		assert(elitismRatio >= 0);
		assert(elitismRatio <= 1);
		assert(reproductionRatio >= 0);
		assert(reproductionRatio <= 1);
		
		// init
		population.resize(populationCount);
		wasLoaded = false;
	}
	
	GenerationalEvolutionRun::~GenerationalEvolutionRun()
	{
		deleteGenomes();
	}
	
	void GenerationalEvolutionRun::deleteGenomes(void)
	{
		std::set<Genome *> genomes;
		
		// collect all genomes
		for (size_t i=0; i<population.size(); i++)
			if (population[i].genome)
				genomes.insert(population[i].genome);
		
		// delete them
		for (std::set<Genome *>::iterator i = genomes.begin(); i != genomes.end(); ++i)
			delete (*i);
	}
	
	double GenerationalEvolutionRun::computeMean(const std::valarray<GenerationalIndividual> &group)
	{
		double sum = 0;
		for (unsigned i = 0; i<group.size(); i++)
			sum += group[i].fitness;
		return sum / group.size();
	}
	
	double GenerationalEvolutionRun::computeSD(const std::valarray<GenerationalIndividual> &group)
	{
		double result = 0;
		double mean = computeMean(group);
		for (size_t i=0; i<group.size(); i++)
		{
			result += (group[i].fitness-mean) * (group[i].fitness-mean);
		}
		result /= group.size();
		result = sqrt(result);
		return result;
	}
	
	void GenerationalEvolutionRun::logPopulation(void)
	{
		double mean = computeMean(population);
		double variance = computeSD(population);
		double diversity = computeDiversity(population);
		double best = -std::numeric_limits<double>::infinity();
		for (size_t i=0; i<population.size(); i++)
			best = std::max<double>(population[i].fitness, best);
	
		std::ofstream summaryLog("fitlog.txt", std::ios_base::out | std::ios_base::app);
		summaryLog << generation << ", " << best << ", " << mean << ", " << variance << ", " << diversity << std::endl;
		
		std::ofstream detailLog("longfitlog.txt", std::ios_base::out | std::ios_base::app);
		detailLog << generation << ", ";
		for (size_t i=0; i<population.size(); i++)
			detailLog << population[i].fitness << ", ";
		detailLog << std::endl;
	}
	
	double GenerationalEvolutionRun::computeDiversity(const std::valarray<GenerationalIndividual> &group)
	{
		std::vector<BitGenome *> bitGenomes(group.size());
		size_t length = 0;
		
		// We assure ourself that we have constant size bitGenome
		for (size_t i=0; i<bitGenomes.size(); i++)
		{
			BitGenome *bg = dynamic_cast<BitGenome *>(group[i].genome);
			if (bg)
			{
				if (length && (length != bg->getLength()))
					return 0;
				bitGenomes[i] = bg;
				length = bg->getLength();
			}
			else
				return 0;
		}
		
		// We compute diversity
		double diversity = 0;
		double groupSize = bitGenomes.size();
		for (size_t i=0; i<length; i++)
		{
			double oneCount = 0;
			for (size_t j=0; j<bitGenomes.size(); j++)
			{
				oneCount += bitGenomes[j]->getBit(i) ? 1 : 0;
			}
			double sumsq = oneCount - ((oneCount * oneCount) / groupSize);
			
			diversity += sumsq / groupSize;
		}
		
		return (diversity * 4) / length;
	}
	
	bool GenerationalEvolutionRun::loadPopulationFromStream(Experiment* experiment, Ishtar::InputStream *stream)
	{
		char buf[4];
		
		// read header
		stream->read(buf, 4);
		if (memcmp(buf, "TEEM", 4) != 0)
			return false;
			
		stream->read(buf, 4);
		if (memcmp(buf, "-POP", 4) != 0)
			return false;
		/*Ishtar::UInt32 teemVersion =*/ stream->readUInt32();
		/*Ishtar::UInt32 fileFormatVersion =*/ stream->readUInt32();
		/*Ishtar::UInt32 flags =*/ stream->readUInt32();
		
		// read datas
		Ishtar::UInt32 individualCount = stream->readUInt32();
		population.resize(individualCount);
		generation = stream->readUInt32();
		
		// load individuals
		for (size_t i=0; i<population.size(); i++)
		    population[i].load(stream);
			
		wasLoaded = true;
		
		return true;
	}
	
	bool GenerationalEvolutionRun::loadPopulation(Experiment* experiment)
	{
		DIR *dir;
		
		dir = opendir(".");
		if (dir)
		{
			unsigned maxGeneration = 0;
			
			struct dirent *entry;
			while ((entry = readdir(dir)) != NULL)
			{
				// TODO : update maxGeneration cleaner
				unsigned filenameGen;
				if (sscanf(entry->d_name, "%04d.pop", &filenameGen)==1)
				{
					maxGeneration = std::max(maxGeneration, filenameGen);
				}
			}
			closedir(dir);
			
			return loadPopulation(experiment, maxGeneration);
		}
		else
			return loadPopulation(experiment, (unsigned)0);
	}
	
	bool GenerationalEvolutionRun::loadPopulation(Experiment* experiment, unsigned generation)
	{
		// create filename
		std::ostringstream oss;
		oss << std::setw(4) << std::setfill('0') << generation;
		oss << ".pop";
		
		// open stream
		Ishtar::FileInputStream stream(oss.str().c_str());
		
		if (stream)
		{
			return loadPopulationFromStream(experiment, &stream);
		}
		else
			return false;
	}
	
	bool GenerationalEvolutionRun::savePopulation(void)
	{
		// create filename
		std::ostringstream oss;
		oss << std::setw(4) << std::setfill('0') << generation;
		oss << ".pop";
		
		// open stream
		Ishtar::FileOutputStream stream(oss.str().c_str());
		
		if (stream)
		{
			// write header
			stream.write("TEEM-POP", 8);
			stream.writeUInt32(0);
			stream.writeUInt32(0);
			stream.writeUInt32(0);
			
			// write data
			stream.writeUInt32(population.size());
			stream.writeUInt32(generation);
			
			// save individuals
			for (size_t i=0; i<population.size(); i++)
				population[i].save(&stream);
			
			return true;
		}
		else
			return false;
	}
	
	void GenerationalEvolutionRun::createRandomPopulation()
	{
		assert(genotypeDecoders.size() > 0);
		
		for (size_t i=0; i<population.size(); i++)
			population[i].genome = genotypeDecoders[0]->createGenome();
	}
	
	void GenerationalEvolutionRun::run(Evaluable *fitness)
	{
		while (generation < generationCount)
		{
			if (!wasLoaded)
			{
				evaluatePopulation(fitness);
				savePopulation();
				logPopulation();
			}
			else
				wasLoaded = false;
			reproducePopulation();
		}
	}
	
	
	void GenerationalEvolutionRun::run(Setupable *experiment)
	{
		Evaluable *fitness = dynamic_cast<Evaluable *>(experiment);
		if (fitness == NULL)
		{
			std::cerr << "GenerationalEvolutionRun::run : error, experiment is not evaluable" << std::endl;
			exit(1);
		}
		
		run(fitness);
	}
	
	void GenerationalEvolutionRun::testBestIndividual(Setupable *experiment)
	{
		Evaluable *fitness = dynamic_cast<Evaluable *>(experiment);
		if (fitness == NULL)
		{
			std::cerr << "GenerationalEvolutionRun::testBestIndividual : error, experiment is not evaluable" << std::endl;
			exit(1);
		}
		
		testBestIndividual(fitness);
	}
	
	void GenerationalEvolutionRun::testIndividual(Setupable *experiment, const char *individual)
	{
		Evaluable *fitness = dynamic_cast<Evaluable *>(experiment);
		if (fitness == NULL)
		{
			std::cerr << "GenerationalEvolutionRun::testIndividual : error, experiment is not evaluable" << std::endl;
			exit(1);
		}
		
		testIndividual(fitness, atoi(individual));
	}
	
	
	
	void SingleDecoderEvolutionRun::createGenotypeDecoders(Setupable *experiment)
	{
		deleteGenotypeDecoders();
		genotypeDecoders.resize(1);
		genotypeDecoders[0] = genotypeDecodersLibrary->createItem(genotypeDecoderName, "teem.nn");
		//FIXME: root should be "teem.controller"
		
		if (experiment)
			experiment->setupDecoders(genotypeDecoders);
	}
	
	SimpleEvolutionRun::SimpleEvolutionRun()
	{
	}
	
	void SimpleEvolutionRun::evaluatePopulation(Evaluable *fitness)
	{
		// we compute the fitness
		for (unsigned i=0; i<population.size(); i++)
		{
			GenerationalPopulation individual(1);
			individual[0] = population[i];
			fitness->compute(genotypeDecoders, individual, generation, i, Evaluable::MODE_EVOLUTION);
			population[i] = individual[0];
		}
	}
	
	void SimpleEvolutionRun::reproducePopulation(void)
	{
		// integer values
		unsigned elitismNBest = static_cast<unsigned>(elitismRatio * population.size());
		unsigned nReproducer = static_cast<unsigned>(reproductionRatio * population.size());
		
		if (!elitismNBest)
			std::cerr << "SimpleEvolutionRun::evolve : no elitism" << std::endl;
		if (nReproducer<2)
			std::cerr << "SimpleEvolutionRun::evolve : not enough reproducers (min 2)" << std::endl;
		
		// We sort
		std::sort(&population[0], &population[population.size()], individualBacksort);
		
		// New population
		GenerationalPopulation newPopulation(population.size());
		
		// We copy best and reset them
		for (unsigned i=0; i<elitismNBest; i++)
		{
			newPopulation[i] = population[i];
			newPopulation[i].reset();
		}
		
		// We fill the rest with new individual
		unsigned crossoverThreshold = static_cast<unsigned>(crossoverProbability*0xFFF);
		for (unsigned i=elitismNBest; i<population.size(); i++)
		{
			unsigned mother = rand() % nReproducer;
			
			GenerationalIndividual newIndividual;
			newIndividual.genome = population[mother].genome->clone();
			
			if ((unsigned)(rand()&0xFFF) < crossoverThreshold)
			{
				// We have crossover
				unsigned father;
				do
				{
					father = (unsigned)rand() % nReproducer;
				}
				while (father == mother);
				
				newIndividual.genome->cross(population[father].genome);
			}
			
			newPopulation[i] = newIndividual;
		}
		
		// We apply mutations and delete old genomes
		for (unsigned i=elitismNBest; i<newPopulation.size(); i++)
		{
			newPopulation[i].genome->mutate();
			delete (population[i].genome);
		}
		
		// We copy back the new population
		std::copy(&newPopulation[0], &newPopulation[newPopulation.size()], &population[0]);
		
		// Increment generation
		generation++;
	}
	
	void SimpleEvolutionRun::testBestIndividual(Evaluable *experiment)
	{
		// find best
		double bestFit = -std::numeric_limits<double>::infinity();
		size_t best = 0;
		for (size_t i=0; i<population.size(); i++)
			if (population[i].fitness > bestFit)
			{
				best = i;
				bestFit = population[i].fitness;
			}
			
		// test it
		testIndividual(experiment, best);
	}
	
	void SimpleEvolutionRun::testIndividual(Evaluable *fitness, size_t number)
	{
		assert(number < population.size());
		
		GenerationalPopulation testPop(1);
		testPop[0] = population[number];
		fitness->compute(genotypeDecoders, testPop, generation, number, Evaluable::MODE_TEST_INDIVIDUAL);
		std::cout << "Fitness of tested individual is " << testPop[0].fitness << std::endl;
		std::ofstream testIndividualLog("testIndividual.txt", std::ios_base::out | std::ios_base::app);
		testIndividualLog << testPop[0].fitness << std::endl;
	}
	
	MultiDecoderEvolutionRun::MultiDecoderEvolutionRun() :
		numGenotypeDecoder("teem.controllers.count", 0)
	{
		assert(numGenotypeDecoder > 0);
		
		for(unsigned i = 0; i < numGenotypeDecoder; i++)
		{
			std::ostringstream oss;
			oss << "teem.controllers." << i << ".genotypeDecoderName";
			genotypeDecoderNames.push_back(Ishtar::Variable<std::string>(oss.str(), "FeedForward"));
		}
	}
	
	void MultiDecoderEvolutionRun::createGenotypeDecoders(Setupable *experiment)
	{
		deleteGenotypeDecoders();
		genotypeDecoders.resize(numGenotypeDecoder);
		for(unsigned i = 0; i < numGenotypeDecoder; i++)
		{
			std::ostringstream oss;
			oss << "teem.controllers." << i;
			genotypeDecoders[i] = genotypeDecodersLibrary->createItem(genotypeDecoderNames[i], oss.str());
		}
		
		if (experiment)
			experiment->setupDecoders(genotypeDecoders);
	}
	
	SimpleMultiGenomeEvolutionRun::SimpleMultiGenomeEvolutionRun()
	{
		// resize population to hold enough genomes for each individual
		population.resize(populationCount * numGenotypeDecoder);
	}
	
	void SimpleMultiGenomeEvolutionRun::createRandomPopulation()
	{
		assert(genotypeDecoders.size() == numGenotypeDecoder);
		
		for (size_t i=0; i<population.size(); i++)
		{
			size_t decoderIndex = i % numGenotypeDecoder;
			population[i].genome = genotypeDecoders[decoderIndex]->createGenome();
		}
	}
	
	void SimpleMultiGenomeEvolutionRun::evaluatePopulation(Evaluable *fitness)
	{
		assert(population.size() == populationCount * numGenotypeDecoder);
		
		// we compute the fitness
		for(unsigned i = 0; i < populationCount; i++)
		{
			GenerationalPopulation individual(numGenotypeDecoder);
			for(size_t j = 0; j < numGenotypeDecoder; j++)
				individual[j] = population[i*numGenotypeDecoder + j];
			fitness->compute(genotypeDecoders, individual, generation, i, Evaluable::MODE_EVOLUTION);
			for(size_t j = 0; j < numGenotypeDecoder; j++)
			{
				population[i*numGenotypeDecoder + j] = individual[j];
				
				// we copy the fitness of the first genome to all the genome, so
				// that they stay together when sorted afterward.
				if(j != 0)
					population[i*numGenotypeDecoder + j].fitness = individual[0].fitness;
			}
		}
	}
	
	void SimpleMultiGenomeEvolutionRun::reproducePopulation(void)
	{
		// NOTE
		// This is a sort of hack in the sense that there are N genome per logical
		// individual. The N genome must keep the same relative order since they are
		// associated with a particular genotype decoder. The variable populationCount
		// has the logical popupulation size, while population.size() is N*populationCount.
		
		assert(population.size() == populationCount * numGenotypeDecoder);
		
		// integer values
		unsigned elitismNBest = static_cast<unsigned>(elitismRatio * populationCount);
		unsigned nReproducer = static_cast<unsigned>(reproductionRatio * populationCount);
		if (!elitismNBest)
			std::cerr << "SimpleEvolutionRun::evolve : no elitism" << std::endl;
		if (nReproducer<2)
			std::cerr << "SimpleEvolutionRun::evolve : not enough reproducers (min 2)" << std::endl;
		
		// We stable_sort, to enforce items of same fitness to keep relative order.
		std::stable_sort(&population[0], &population[population.size()], individualBacksort);
		
		// New population
		GenerationalPopulation newPopulation(population.size());
		
		// We copy best and reset them
		for (unsigned i=0; i<elitismNBest*numGenotypeDecoder; i++)
		{
			newPopulation[i] = population[i];
			newPopulation[i].reset();
		}
		
		// We fill the rest with new individual
		unsigned crossoverThreshold = static_cast<unsigned>(crossoverProbability*0xFFF);
		for (unsigned i=elitismNBest; i<populationCount; i++)
		{
			unsigned mother = (unsigned)rand() % nReproducer;
			unsigned father = mother;
			bool cross = ((unsigned)rand()&0xFFF) < crossoverThreshold;
			if (cross)
			{
				// We have crossover
				do
				{
					father = rand() % nReproducer;
				}
				while (father == mother);
			}
			
			for(unsigned j = 0; j < numGenotypeDecoder; j++)
			{
				GenerationalIndividual newIndividual;
				newIndividual.genome = population[mother*numGenotypeDecoder+j].genome->clone();
				if (cross)
					newIndividual.genome->cross(population[father*numGenotypeDecoder+j].genome);
				newPopulation[i*numGenotypeDecoder+j] = newIndividual;
			}
		}
		
		// We apply mutations and delete old genomes
		for (unsigned i=elitismNBest*numGenotypeDecoder; i<population.size(); i++)
		{
			newPopulation[i].genome->mutate();
			delete (population[i].genome);
		}
		
		// We copy back the new population
		std::copy(&newPopulation[0], &newPopulation[newPopulation.size()], &population[0]);
		
		// Increment generation
		generation++;
	}
	
	void SimpleMultiGenomeEvolutionRun::testBestIndividual(Evaluable *experiment)
	{
		// find best
		double bestFit = -std::numeric_limits<double>::infinity();
		size_t best = 0;
		for (size_t i=0; i<populationCount; i++)
			if (population[i*numGenotypeDecoder].fitness > bestFit)
			{
				bestFit = population[i*numGenotypeDecoder].fitness;
				best = i;
			}
			
		// test it
		testIndividual(experiment, best);
	}
	
	void SimpleMultiGenomeEvolutionRun::testIndividual(Evaluable *fitness, size_t number)
	{
		assert(number < populationCount);
		
		GenerationalPopulation testPop(numGenotypeDecoder);
		
		for(unsigned i = 0; i < numGenotypeDecoder; i++)
			testPop[i] = population[number * numGenotypeDecoder + i];
		
		fitness->compute(genotypeDecoders, testPop, generation, number, Evaluable::MODE_TEST_INDIVIDUAL);

		std::cout << "Fitness of tested individual is " << testPop[0].fitness << std::endl;
		std::ofstream testIndividualLog("testIndividual.txt", std::ios_base::out | std::ios_base::app);
		testIndividualLog << testPop[0].fitness << std::endl;
	}
}
