/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "SimpleGrowableGenome.h"
#include <an/Random.h>
#include <algorithm>
#include <set>
#include <map>
#include <cmath>
#include <assert.h>

/*!	\file SimpleGrowableGenome.cpp
	\brief Implementation of genomes based on graphs of nodes and weighted links
*/

namespace Teem
{
	using namespace An;

	SimpleGrowableGenome::Link::Link(unsigned in, unsigned out, double weight)
	{
		this->in = in;
		this->out = out;
		assert(finite(weight));
		this->weight = weight;
	}
	
	SimpleGrowableGenome::Link::Link(Ishtar::InputStream *stream)
	{
		in = stream->readUInt32();
		out = stream->readUInt32();
		weight = stream->readDouble();
	}
	
	void SimpleGrowableGenome::Link::save(Ishtar::OutputStream *stream)
	{
		stream->writeUInt32(in);
		stream->writeUInt32(out);
		stream->writeDouble(weight);
	}

	
	SimpleGrowableGenome::SimpleGrowableGenome(Ishtar::InputStream *stream) :
		mutationAddLinkConstProbability("teem.ga.mutationAddLinkConstProbability", 0.1),
		mutationAddLinkPropProbability("teem.ga.mutationAddLinkPropProbability", 0.01),
		mutationDelLinkProbability("teem.ga.mutationDelLinkProbability", 0.02),
		mutationChangeLinkProbability("teem.ga.mutationChangeLinkProbability", 0.03),
		mutationAddNodeProbability("teem.ga.mutationAddNodeProbability", 0.01),
		mutationDelNodeProbability("teem.ga.mutationDelNodeProbability", 0.01),
		mutationDuplicateGenomeProbability("teem.ga.mutationDuplicateGenomeProbability", 0.001)
	{
		error=false;
		minimumToPreserve = stream->readUInt32();
		size_t linkCount = stream->readUInt32();
		for (size_t i = 0; i < linkCount; i++)
			links.push_back(Link(stream));
	}
	
	SimpleGrowableGenome::SimpleGrowableGenome(unsigned minToPreserve) :
		mutationAddLinkConstProbability("teem.ga.mutationAddLinkConstProbability", 0.1),
		mutationAddLinkPropProbability("teem.ga.mutationAddLinkPropProbability", 0.01),
		mutationDelLinkProbability("teem.ga.mutationDelLinkProbability", 0.02),
		mutationChangeLinkProbability("teem.ga.mutationChangeLinkProbability", 0.03),
		mutationAddNodeProbability("teem.ga.mutationAddNodeProbability", 0.01),
		mutationDelNodeProbability("teem.ga.mutationDelNodeProbability", 0.01),
		mutationDuplicateGenomeProbability("teem.ga.mutationDuplicateGenomeProbability", 0.001)
	{
		error=false;
		minimumToPreserve = minToPreserve;
	}
	
	unsigned SimpleGrowableGenome::countNodes(void)
	{
		unsigned count = 0;
		for (std::list<Link>::iterator it = links.begin(); it != links.end(); ++it)
		{
			count = std::max(count, it->in);
			count = std::max(count, it->out);
		}
		return count;
	}
		
	void SimpleGrowableGenome::defragNodes(void)
	{
		std::set<unsigned> usedNodes;
		
		// collect used nodes
		for (std::list<Link>::iterator it = links.begin(); it != links.end(); ++it)
		{
			usedNodes.insert(it->in);
			usedNodes.insert(it->out);
		}
		
		std::map<unsigned, unsigned> nodeIdTranslationTable;
		// build translation table
		std::set<unsigned>::iterator toDefragIt;
		if (minimumToPreserve > 0)
			toDefragIt = usedNodes.upper_bound(minimumToPreserve - 1);
		else
			toDefragIt = usedNodes.begin();
		unsigned newId = minimumToPreserve;
		while (toDefragIt != usedNodes.end())
			nodeIdTranslationTable[*toDefragIt++] = newId++;
		
		// translate
		for (std::list<Link>::iterator it = links.begin(); it != links.end(); ++it)
		{
			if (nodeIdTranslationTable.find(it->in) != nodeIdTranslationTable.end())
				it->in = nodeIdTranslationTable[it->in];
			if (nodeIdTranslationTable.find(it->out) != nodeIdTranslationTable.end())
				it->out = nodeIdTranslationTable[it->out];
		}
	}
	
	void SimpleGrowableGenome::mutateAddLink(unsigned nodeCount)
	{
		unsigned possibleNodes = std::max(nodeCount, minimumToPreserve);
		assert(possibleNodes > 0);
		if (uniformRand() < mutationAddLinkConstProbability)
			links.push_back(Link(intRand(possibleNodes), intRand(possibleNodes), gaussianRand(0, 0.5)));
		for (size_t i=0; i<links.size(); i++)
			if (uniformRand() < mutationAddLinkPropProbability)
				links.push_back(Link(intRand(possibleNodes), intRand(possibleNodes), gaussianRand(0, 0.5)));
	}
	
	void SimpleGrowableGenome::mutateDelLink(void)
	{
		for (std::list<Link>::iterator it = links.begin(); it != links.end();)
			if (uniformRand() < mutationDelLinkProbability)
				it = links.erase(it);
			else
				++it;
	}
	
	void SimpleGrowableGenome::mutateChangeLink(void)
	{
		for (std::list<Link>::iterator it = links.begin(); it != links.end(); ++it)
			if (uniformRand() < mutationChangeLinkProbability)
				it->weight += gaussianRand(0, 0.3);
	}
	
	void SimpleGrowableGenome::mutateAddNode(unsigned nodeCount)
	{
		unsigned newNodeId = nodeCount;
		for (std::list<Link>::iterator it = links.begin(); it != links.end(); ++it)
			if (uniformRand() < mutationAddNodeProbability)
			{
				double nw0, nw1;
				if (it->weight > 0)
				{
					nw0 = nw1 = sqrt(it->weight);
				}
				else
				{
					nw0 = sqrt(-it->weight);
					nw1 = -nw0;
				}
				links.push_back(Link(it->in, newNodeId, nw0));
				*it = Link(newNodeId, it->out, nw1);
				newNodeId++;
			}
	}
	
	void SimpleGrowableGenome::mutateDelNode(unsigned nodeCount)
	{
		for (unsigned i=0; i<nodeCount; i++)
			if (uniformRand() < mutationDelNodeProbability)
				for (std::list<Link>::iterator it = links.begin(); it != links.end();)
					if ((it->in == i) || (it->out == i))
						it = links.erase(it);
					else
						++it;
	}
	
	void SimpleGrowableGenome::mutateDuplicateGenome(void)
	{
		if (uniformRand() < mutationDuplicateGenomeProbability)
		{
			size_t l = links.size();
			size_t i = 0;
			std::list<Link>::const_iterator it = links.begin();
			while (i++ < l)
				links.push_back(*it++);
		}
	}
	
	void SimpleGrowableGenome::randomize(void)
	{
		assert(minimumToPreserve > 0);
		links.push_back(Link(intRand(minimumToPreserve), intRand(minimumToPreserve), gaussianRand(0, 0.5)));
	}
	
	void SimpleGrowableGenome::mutate(void)
	{
		unsigned nodeCount = countNodes();
		
		mutateAddLink(nodeCount);
		mutateDelLink();
		
		mutateAddNode(nodeCount);
		mutateDelNode(nodeCount);

		mutateDuplicateGenome();
		
		defragNodes();
	}
	
	void SimpleGrowableGenome::cross(const Genome *other)
	{
		// No cross for now, because we need unique id not yet available
	}
	
	Genome *SimpleGrowableGenome::clone(void)
	{
		return new SimpleGrowableGenome(*this);
	}
	
	void SimpleGrowableGenome::save(Ishtar::OutputStream *stream)
	{
		stream->writeUInt32(static_cast<Ishtar::UInt32>(SIMPLE_GROWABLE_GENOME));
		stream->writeUInt32(minimumToPreserve);
		stream->writeUInt32(links.size());
		for (std::list<Link>::iterator it = links.begin(); it != links.end(); ++it)
			it->save(stream);
	}
}
