/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef __SIMPLE_GROWABLE_GENOME_H
#define __SIMPLE_GROWABLE_GENOME_H

#include "Genome.h"
#include <list>

/*!	\file SimpleGrowableGenome.h
	\brief Interfaces of genomes based on graphs of nodes and weighted links
*/

namespace Teem
{
	//! A genome that represent a graph with nodes and weightened links that can grow and shrink
	//! \ingroup geneticalgos
	class SimpleGrowableGenome: public Genome
	{
	public:
		//! A link
		struct Link
		{
			unsigned in; //!< Source node id
			unsigned out; //!< Destination node id
			double weight; //!< Weight of this link
			
			//! Constructor, link in to out using weight
			Link(unsigned in, unsigned out, double weight);
			//! Constructor, read in, out and weight from stream
			Link(Ishtar::InputStream *stream);
			//! Save in, out and weight to stream
			void save(Ishtar::OutputStream *stream);
		};
	
	protected:
		friend class NeatNeuralNetwork;
		//static unsigned innovationCount;
		std::list<Link> links; //!< all links in the genome
		unsigned minimumToPreserve; //!< minimum number of node id to preserve on shrink because they are I/O nodes
		
	protected:
		// Support function
		//! Return the number of nodes (i.e. the index of the node with biggest index)
		unsigned countNodes(void); 
		//! Defrag the node index space by reusing unused index. Do not compress nodes below minimumToPreserve
		void defragNodes(void);
		
		// Mutation operator
		//! Add new links at random between two nodes with index up to nodeCount
		void mutateAddLink(unsigned nodeCount);
		//! Remove links at random
		void mutateDelLink(void);
		//! Change the value of links by adding to it a random value following a Gaussian distribution of mean 0 and standard deviation 0.3
		void mutateChangeLink(void);
		
		//! Add new nodes by cutting links in two starting from index at nodeCount 
		void mutateAddNode(unsigned nodeCount);
		//! Delete nodes at random with index up to nodeCount
		void mutateDelNode(unsigned nodeCount);
		
		//! Duplicate the whole genome
		void mutateDuplicateGenome(void);
		
		// Ishtar parameters
		Ishtar::Variable<double> mutationAddLinkConstProbability; //!< probability of adding a new link at random
		Ishtar::Variable<double> mutationAddLinkPropProbability; //!< probability of adding a new link by cutting an existing link in two
		Ishtar::Variable<double> mutationDelLinkProbability; //!< probability of deleting a link
		Ishtar::Variable<double> mutationChangeLinkProbability; //!< probability of changing the value of a link
		
		Ishtar::Variable<double> mutationAddNodeProbability; //!< probability of adding a new node
		Ishtar::Variable<double> mutationDelNodeProbability; //!< probability of deleting an existing node
		
		Ishtar::Variable<double> mutationDuplicateGenomeProbability; //!< probability of duplicating the whole genome
	
	public:
		//! Constructor, reading genome from stream
		SimpleGrowableGenome(Ishtar::InputStream *stream);
		//! Constructor, minToPreserve is the first n node that cannot be defragmented because they are used for I/O
		SimpleGrowableGenome(unsigned minToPreserve);
		
		//! Destructor
		virtual ~SimpleGrowableGenome() { }
		virtual void randomize(void);
		virtual void mutate(void);
		virtual void cross(const Genome *other);
		virtual Genome *clone(void);
		
		virtual void save(Ishtar::OutputStream *stream);
	};
}

#endif
