/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "SteadyStateEvolution.h"
#include "../ga/Genome.h"
#include "../controller/GenotypeDecoder.h"
#include "../experiment/Experiment.h"
#include "../main/Registry.h"
#include <ishtar/network.h>
#include <ishtar/filestream.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <ostream>
#include <iomanip>
#include <limits>
#include <dirent.h>

/*!	\file SteadyStateEvolution.cpp
	\brief Implementation of algorithms for steady-state (continuous) evolution
*/

namespace Teem
{
    // SteadyStateIndividual
	SteadyStateIndividual::SteadyStateIndividual(Genome *genome, double bornTime)
	{
		this->genome = genome;
		this->bornTime = bornTime;
		deathTime = 0;
		childrenCount = 0;
		userData = NULL;
	}
	
	SteadyStateIndividual::SteadyStateIndividual(GenotypeDecoders &decoders, Continuable *experiment, Ishtar::InputStream *stream)
	{
		genome = NULL;
		userData = NULL;
		load(decoders, experiment, stream);
	}
	
	SteadyStateIndividual::~SteadyStateIndividual()
	{
		if (genome)
			delete genome;
		if (userData)
			delete userData;
	}
	
	void SteadyStateIndividual::load(GenotypeDecoders &decoders, Continuable *experiment, Ishtar::InputStream *stream)
	{
		bornTime = stream->readDouble ();
		deathTime = stream->readDouble ();
		childrenCount = stream->readUInt32 ();
		if (genome)
			delete genome;
		genome = Genome::load(stream);
		if (userData)
			delete userData;
		userData = experiment->individualCreated(decoders, this, NULL, NULL);
		if (userData)
			userData->load(stream);
	}
	
	void SteadyStateIndividual::save(Ishtar::OutputStream *stream)
	{
		stream->writeDouble(bornTime);
		stream->writeDouble(deathTime);
		stream->writeUInt32(childrenCount);
		genome->save(stream);
		if (userData)
			userData->save(stream);
	}
	
	
	// SteadyStateEvolutionRun
	SteadyStateEvolutionRun::SteadyStateEvolutionRun() :
	initialPopulationSize("teem.ga.initialPopulationSize", 100),
	evolutionDuration("teem.ga.evolutionDuration", 3600),
	genotypeDecoderName("teem.genotypeDecoderName", "FeedForward"),
	timeSaveInterval("teem.ga.timeSaveInterval", 100.0)
	{
		time = 0.0;
		nextSaveTime = 0.0;
	}
	
	SteadyStateEvolutionRun::~SteadyStateEvolutionRun()
	{
		// delete all individuals
		for (SteadyStatePopulation::iterator i = population.begin(); i != population.end(); ++i)
			delete *i;
		assert(toCreate.size() == 0);
		assert(toKill.size() == 0);
	}
	
	void SteadyStateEvolutionRun::createGenotypeDecoders(Setupable *experiment)
	{
		deleteGenotypeDecoders();
		genotypeDecoders.resize(1);
		genotypeDecoders[0] = genotypeDecodersLibrary->createItem(genotypeDecoderName, "teem.controller");
			
		if (experiment)
			experiment->setupDecoders(genotypeDecoders);
	}
	
	void SteadyStateEvolutionRun::createRandomPopulation()
	{
		assert(initialPopulationSize > 0);
			
		for (unsigned i=0; i<initialPopulationSize; i++)
			toCreate.insert(NewbornIndividuals(new SteadyStateIndividual(genotypeDecoders[0]->createGenome(), time), NULL, NULL));
	}
	
	bool SteadyStateEvolutionRun::loadPopulation(Experiment *experiment)
	{
		DIR *dir = opendir(".");
		assert(dir);
		if (dir)
		{
			double maxTime = 0;
			
			struct dirent *entry;
			while ((entry = readdir(dir)) != NULL)
			{
				double filenameTime;
				if (sscanf(entry->d_name, "%lf.pop", &filenameTime) == 1)
				{
					maxTime = std::max(maxTime, filenameTime);
				}
			}
			closedir(dir);
			return loadPopulation(experiment, maxTime);
		}
		else
			return false;
	}
	
	bool SteadyStateEvolutionRun::loadPopulation(Experiment *experiment, double timeToLoad)
	{
		// create filename
		char filename[80];
		sprintf(filename, "%0.2lf.pop", timeToLoad);
	
		// open stream
		Ishtar::FileInputStream stream(filename);
		
		assert(stream);
		if (stream)
			return loadPopulationFromStream(experiment, &stream);
		else
			return false;
	}
	
	
	bool SteadyStateEvolutionRun::loadPopulationFromStream(Experiment *experiment, Ishtar::InputStream *stream)
	{
		Continuable* continuableExperiment = dynamic_cast<Continuable *>(experiment);
		if (continuableExperiment == NULL)
		{
			std::cerr << "SteadyStateEvolutionRun::run : error, experiment is not steady state." << std::endl;
			exit(1);
		}
		
		char buf[4];
		
		// read header
		stream->read(buf, 4);
		if (memcmp(buf, "TEEM", 4) != 0)
			return false;
		
		stream->read(buf, 4);
		if (memcmp(buf, "-POP", 4) != 0)
			return false;
		/*Ishtar::UInt32 teemVersion =*/ stream->readUInt32();
		/*Ishtar::UInt32 fileFormatVersion =*/ stream->readUInt32();
		/*Ishtar::UInt32 flags =*/ stream->readUInt32();
			
		// read datas
		Ishtar::UInt32 individualCount = stream->readUInt32();
		time = stream->readDouble();
		nextSaveTime = time + timeSaveInterval;
		
		// read experiment data
		Serializable *serializableExperiment = dynamic_cast<Serializable *>(experiment);
		if (serializableExperiment)
			serializableExperiment->load(stream);
		
		// first delete all individuals
		for (SteadyStatePopulation::iterator i = population.begin(); i != population.end(); ++i)
			delete *i;
		toCreate.clear();
		for (SteadyStatePopulation::iterator i = toKill.begin(); i != toKill.end(); ++i)
			delete *i;
		toKill.clear();
		
		// load individuals
		for (unsigned i=0 ; i != individualCount ; i++)
			population.insert(new SteadyStateIndividual(genotypeDecoders, continuableExperiment, stream));
		
		return true;
	}
	
	
	bool SteadyStateEvolutionRun::savePopulation(Setupable *experiment)
	{
		// create filename
		char filename[80];
		sprintf(filename, "%0.2lf.pop", time);
		
		// open stream
		Ishtar::FileOutputStream stream(filename);
		
		if (stream)
		{
			// write header
			stream.write("TEEM-POP", 8);
			stream.writeUInt32(0);
			stream.writeUInt32(0);
			stream.writeUInt32(0);
				
			// write data
			stream.writeUInt32(population.size());
			stream.writeDouble(time);
		
			// write experiment data
			Serializable *serializableExperiment = dynamic_cast<Serializable *>(experiment);
			if (serializableExperiment)
				serializableExperiment->save(&stream);
			
			// save individuals 
			for (SteadyStatePopulation::iterator i = population.begin(); i != population.end(); ++i)
				(*i)->save(&stream);
			
			return true;
		}
		else
			return false;
	}
	
	void SteadyStateEvolutionRun::run(Setupable *experiment)
	{
		Continuable *continuable = dynamic_cast<Continuable *>(experiment);
		if (continuable == NULL)
		{
			std::cerr << "SteadyStateEvolutionRun::run : error, experiment is not steady state." << std::endl;
			exit(1);
		}
		
		while (time < evolutionDuration)
		{
			// step
			time += continuable->step(this, time);
			
			// create new individuals
			for (NewbornPopulation::iterator i = toCreate.begin(); i != toCreate.end(); ++i)
			{
				i->self->userData = continuable->individualCreated(genotypeDecoders, i->self, i->mother, i->father);
				population.insert(i->self);
			}
			toCreate.clear();
			
			// kill individuals
			for (SteadyStatePopulation::iterator i = toKill.begin(); i != toKill.end(); ++i)
			{
				population.erase(*i);
				delete *i;
			}
			toKill.clear();
			
			// save / log population
			if (time >= nextSaveTime)
			{
				nextSaveTime += timeSaveInterval;
				std::ofstream summaryLog("populationlog.txt", std::ios_base::out | std::ios_base::app);
				double totalLife = 0;
				double totalChildrenCount = 0;
				double count = 0;
				double maxLifeTime = 0;
				for (SteadyStatePopulation::iterator i = population.begin(); i != population.end(); ++i)
				{
					double lifeTime = time - (*i)->bornTime;
					maxLifeTime = std::max(maxLifeTime, lifeTime);
					totalLife += lifeTime;
					totalChildrenCount += (*i)->childrenCount;
					count ++;
				}
				summaryLog << time << ", " << population.size() << ", " << maxLifeTime << ", " << (totalLife / count) << ", " << (totalChildrenCount / count) << std::endl;
				savePopulation(experiment);
			}
		}
	}
	
	void SteadyStateEvolutionRun::testBestIndividual(Setupable *experiment)
	{
		Continuable *continuableExperiment = dynamic_cast<Continuable *>(experiment);
		if (continuableExperiment == NULL)
		{
			std::cerr << "SteadyStateEvolutionRun::run : error, experiment is not steady state." << std::endl;
			exit(1);
		}
		
		while (true)
			continuableExperiment->step(this, 0);
	}
	
	void SteadyStateEvolutionRun::testIndividual(Setupable *experiment, const char* testTimeString)
	{
		Continuable *continuableExperiment = dynamic_cast<Continuable *>(experiment);
		if (continuableExperiment == NULL)
		{
			std::cerr << "SteadyStateEvolutionRun::run : error, experiment is not steady state." << std::endl;
			exit(1);
		}
		
		double testTime = atof(testTimeString);
		
		while (true)
			continuableExperiment->step(this, testTime);
	}
	
	void SteadyStateEvolutionRun::createRandomIndividual()
	{
		toCreate.insert(NewbornIndividuals(new SteadyStateIndividual(genotypeDecoders[0]->createGenome(), time), NULL, NULL));
	}
	
	void SteadyStateEvolutionRun::createIndividual(SteadyStateIndividual *parent)
	{
		parent->childrenCount++;
		Genome * newGenome = parent->genome->clone();
		newGenome->mutate();
		toCreate.insert(NewbornIndividuals(new SteadyStateIndividual(newGenome, time), parent, NULL));
	}
	
	void SteadyStateEvolutionRun::createIndividual(SteadyStateIndividual *mother, SteadyStateIndividual *father)
	{
		mother->childrenCount++;
		father->childrenCount++;
		Genome * newGenome = mother->genome->clone();
		newGenome->cross(father->genome);
		newGenome->mutate();
		toCreate.insert(NewbornIndividuals(new SteadyStateIndividual(newGenome, time), mother, father));
	}
	
	void SteadyStateEvolutionRun::killIndividual(SteadyStateIndividual *individual)
	{
		toKill.insert(individual);
	}
}
