/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors.
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef STEADY_STATE_EVOLUTION_H
#define STEADY_STATE_EVOLUTION_H

#include "Evolution.h"
#include <set>
#include <ishtar/settings.h>

/*!	\file SteadyStateEvolution.h
	\brief Interfaces of algorithms for steady-state (continuous) evolution
*/

namespace Ishtar
{
	class InputStream;
	class OutputStream;
}

namespace Teem
{
	class Genome;
	
	class SteadyStateEvolutionRun;
	class Continuable;
	
	//! A individual for steady-state evolution
	//! \ingroup geneticalgos
	class SteadyStateIndividual
	{
	public:
		//! The genetic material for this individual
		Genome *genome;
		//! The time this individual was born
		double bornTime;
		//! The time this individual died
		double deathTime;
		//! The number of children this individual had
		unsigned childrenCount;
		//! This user data contains all experiment-related stuff for this individual. If not NULL, it is delete when individual is killed
		Serializable *userData;
		
	public:
		//! Constructor
		SteadyStateIndividual(Genome *genome, double bornTime);
		//! Constructor loading an individual from a stream, and calling the experiment individualCreated method with read datas
		SteadyStateIndividual(GenotypeDecoders &decoders, Continuable *experiment, Ishtar::InputStream *stream);
		//! Destructor
		~SteadyStateIndividual();
		
		//! Load this individual's timestamps, genome and user datas
		void load(GenotypeDecoders &decoders, Continuable *experiment, Ishtar::InputStream *stream);
		//! Save this indivitual's timestamps, genome and user datas
		void save(Ishtar::OutputStream *stream);
	};
	
	//! Population of individual suited for steady-state evolution
	//! \ingroup geneticalgos
	typedef std::set<SteadyStateIndividual *> SteadyStatePopulation;
	
	//! Interface to a class that implements a continuous evolution.
	//! Experiments classes for steady-state evolution typically derive from Continuable.
	//! \ingroup geneticalgos
	class Continuable : public Setupable
	{
	public:
		//! Do a step of continuous evolution. The elapsed time is returned
		virtual double step(SteadyStateEvolutionRun *evolution, double time) = 0;
		//! A new individual has been created. This function normally returns the user datas associated with this individual. Nevertheless, it may return NULL if there is no user data
		virtual Serializable *individualCreated(GenotypeDecoders &decoders, SteadyStateIndividual *self, SteadyStateIndividual *mother, SteadyStateIndividual *father) = 0;
	};
	
	//! The basic class for steady-state evolution
	//! \ingroup geneticalgos
	class SteadyStateEvolutionRun : public EvolutionRun
	{
	protected:
		// inner class
		
		//! A newly born individual, keep tracks of father and mother
		struct NewbornIndividuals
		{
			//! Constructor
			NewbornIndividuals(SteadyStateIndividual *self, SteadyStateIndividual *mother, SteadyStateIndividual *father) { this->self = self; this->mother = mother; this->father = father; }
			
			SteadyStateIndividual *self; //!< pointer to self
			SteadyStateIndividual *mother; //!< pointer to mother
			SteadyStateIndividual *father; //!< pointer to father
		};
		
		//! Function for NewbornIndividual sorting in set
		struct NewbornIndividualsCompareFunctor
		{
			//! () operator, do the comparaison
			bool operator()(const NewbornIndividuals &i1, const NewbornIndividuals &i2) const
			{
				return i1.self < i2.self;
			}
		};
		
		//! Population of newborn individual, keep track of father and mother
		typedef std::set<NewbornIndividuals, NewbornIndividualsCompareFunctor> NewbornPopulation;
		
	protected:
		// members
		//! The initial number of individuals present in the population
		Ishtar::Variable<unsigned> initialPopulationSize;
		//! The duration of the evolution
		Ishtar::Variable<double> evolutionDuration;
		//! Name of the genotype decoder to be used : will be perhaps moved at another place later
		Ishtar::Variable<std::string> genotypeDecoderName;
		//! Interval in s of simulated time between two saves
		Ishtar::Variable<double> timeSaveInterval;
		
		//! A temporary population of individual that will be born before next step
		NewbornPopulation toCreate;
		//! A temporary population of individual that will be killed before next step
		SteadyStatePopulation toKill;
		//! The actual time
		double time;
		//! The next time after which we have to save
		double nextSaveTime;
		
	public:
		//! A population, consisting of all individuals
		SteadyStatePopulation population;
	
	protected:
		// protected methods used internally
		//! Save the current population to the file
		virtual bool savePopulation(Setupable *experiment);
		//! Load a population from a specific generation or time
		virtual bool loadPopulation(Experiment *experiment, double timeToLoad);
		//! Get the population through Ishtar
		virtual bool loadPopulationFromStream(Experiment *experiment, Ishtar::InputStream *stream);
	
	public:
		// public methods from Evolution
		//! Constuctor
		SteadyStateEvolutionRun();
		//! Destructor
		virtual ~SteadyStateEvolutionRun();
		
		//! Create the necessary genotype decoders.
		virtual void createGenotypeDecoders(Setupable *experiment);
		
		//! Create a population of random genomes, using the genotype decoder
		virtual void createRandomPopulation();

		//! Load a most recent population
		virtual bool loadPopulation(Experiment *experiment);
		//! Load a population at a certain time
		virtual bool loadPopulation(Experiment *experiment, const char *timeToLoad) { return loadPopulation(experiment, atof(timeToLoad)); }
		
		//! Do the evolution
		virtual void run(Setupable *experiment);
			
		//! Test the best Individual and test it (used to view and evaluate indivuals for analysis)
		virtual void testBestIndividual(Setupable *experiment);
		//! Test a certain individual and evaluate it (used to view and evaluate individuals for analysis)
		virtual void testIndividual(Setupable *experiment, const char *);
		
	public:
		// public methods to be called from Continuable
		//! Create a new random individual
		virtual void createRandomIndividual();
		//! Create a new individual from a single parent
		virtual void createIndividual(SteadyStateIndividual *parent);
		//! Create a new individual from a single parent
		virtual void createIndividual(SteadyStateIndividual *mother, SteadyStateIndividual *father);
		//! Kill an individual
		virtual void killIndividual(SteadyStateIndividual *individual);
	};
}

#endif
