/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "ValueGenome.h"
#include <assert.h>

/*!	\file ValueGenome.cpp
	\brief Implementations of genomes based on array of values
*/

namespace Teem
{
	BitGenome::BitGenome(Ishtar::InputStream *stream) :
		mutationProbability("teem.ga.mutationProbability", 0.01),
		defaultPrecision("teem.ga.defaultPrecision", 8)
	{
		error = false;

		length = stream->readUInt32();
		size_t wordCount = (length / 32) + 1;
		bits.resize(wordCount);
	
		// init data
		for (size_t i=0; i<bits.size(); i++)
			bits[i] = stream->readUInt32();
			
		// integrity constraints
		assert(mutationProbability >= 0);
		assert(mutationProbability <= 1);
	
		// init internal machinery
		readPos = 0;
	}
	
	BitGenome::BitGenome(size_t size) :
		mutationProbability("teem.ga.mutationProbability", 0.01),
		defaultPrecision("teem.ga.defaultPrecision", 8)
	{
		error = false;

		// only works when sizeof(unsigned) == 4
		assert(sizeof(unsigned) == 4);
		
		// init data
		size_t wordCount = (size / 32) + 1;
		bits.resize(wordCount);
		length = size;
		
		// integrity constraints
		assert(mutationProbability >= 0);
		assert(mutationProbability <= 1);
	
		// init internal machinery
		readPos = 0;
	}
	
	bool BitGenome::checkBoundWord(size_t pos)
	{
		if (pos < bits.size())
			return true;
		
		error = true;
		if (!disableAssertion)
		{
			std::cerr << "BitGenome::checkBoundWord : index " << pos << " out of bounds " << bits.size() << std::endl;
			assert(false);
		}
		
		return false;
	}
	
	// disable because currently not used
	/*bool BitGenome::checkBoundBit(size_t pos)
	{
		if (pos < length)
			return true;
		
		error = true;
		if (!disableAssertion)
		{
			std::cerr << "BitGenome::checkBoundBit : index " << pos << " out of bounds " << length << std::endl;
			assert(false);
		}
		
		return false;
	}*/
	
	bool BitGenome::getBit(size_t pos)
	{
		size_t wordPos = pos / 32;
		size_t bitPos = pos % 32;
	
		if (checkBoundWord(wordPos))
			return (bits[wordPos] & (1<<bitPos));
		else
			return false;
	}
	
	void BitGenome::setBit(size_t pos, bool bit)
	{
		size_t wordPos = pos / 32;
		size_t bitPos = pos % 32;
		
		if (checkBoundWord(wordPos))
		{
			if (bit)
				bits[wordPos] |= (1<<bitPos);
			else
				bits[wordPos] &= ~(1<<bitPos);
		}
	}
	
	void BitGenome::seek(size_t pos)
	{
		assert(pos < length);
		readPos = pos;
	}
	
	void BitGenome::resetReading(void)
	{
		if (readPos != length)
			std::cerr << "BitGenome::resetReading() : reading position reseted before end of genome" << std::endl;
		readPos = 0;
	}
	
	void BitGenome::randomize(void)
	{
		for (size_t i=0; i<bits.size(); i++)
		{
			bits[i] = 0;
			for (unsigned j=0; j<4; j++)
				bits[i] |= (rand()&0xFF) << (j*8);
		}
	}
	
	void BitGenome::mutate(void)
	{
		unsigned mutationThreshold = static_cast<unsigned>(mutationProbability*0xFFF);
		for (size_t i=0; i<bits.size(); i++)
		{
			for (unsigned j=0; j<32; j++)
			{
				if ((unsigned)(rand()&0xFFF) < mutationThreshold)
					bits[i] ^= (1<<j);
			}
		}
	}
	
	void BitGenome::cross(const Genome *other)
	{
		const BitGenome *obg = dynamic_cast<const BitGenome *>(other);
		if (obg)
		{
			if (length != obg->length)
			{
				error = true;
				if (disableAssertion)
					return;
				else
					assert(false);
			}
			
			size_t crossPos = rand()%length;
			size_t crossByte = crossPos >> 5;
			unsigned crossBit = crossPos & 0x1F;
		
			unsigned v0 = bits[crossByte];
			unsigned v1 = obg->bits[crossByte];
		
			unsigned mask = (1<<crossBit)-1;
			unsigned nmask = ~mask;
		
			unsigned mnv = (v0 & mask) | (v1 & nmask);
		
			bits[crossByte] = mnv;
			
			for (size_t i=crossByte+1; i<bits.size(); i++)
			{
				bits[i] = obg->bits[i];
			}
		}
		else
		{
			std::cerr << "BitGenome::cross() : trying to cross with a non-bit genome, ignored" << std::endl;
		}
	}
	
	Genome *BitGenome::clone(void)
	{
		return new BitGenome(*this);
	}
	
	void BitGenome::save(Ishtar::OutputStream *stream)
	{
		stream->writeUInt32(static_cast<Ishtar::UInt32>(BIT_GENOME));
		stream->writeUInt32(length);
		for (size_t i=0; i<bits.size(); i++)
			stream->writeUInt32(bits[i]);
	}
}
