/*
  Copyright (C) 2001-2004 Stephane Magnenat & Luc-Olivier de Charrière
  for any question or comment contact us at nct@ysagoon.com or nuage@ysagoon.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef __GRAPHICCONTEXT_H
#define __GRAPHICCONTEXT_H

#include "GAGSys.h"
#include <map>
#include <vector>

/*!	\file GraphicContext.h
	\brief Interface of basic drawing primitives
*/

namespace GAG
{
	class Font;
	class Sprite;
	
	//! A surface of pixel that can be drawn on
	//! \ingroup libgag
	class DrawableSurface
	{
	public:
		//! Specific alpha values
		enum Alpha
		{
			ALPHA_TRANSPARENT=0, //!< transparent
			ALPHA_OPAQUE=255 //!< opaque
		};
	
		//! Flags for surface
		enum ResolutionFlags
		{
			DEFAULT=0, //!< no flags
			DOUBLEBUF=1, //!< surface is double-buffered
			FULLSCREEN=2, //!< surface is fullscreen
			HWACCELERATED=4, //!< surface is stored in graphic card
			RESIZABLE=8 //!< surface can be resized
		};
	
	public:
		//! Virtual destructor, do nothing
		virtual ~DrawableSurface(void) { }
		//! Set the resolution
		virtual bool setRes(int w, int h, int depth=32, Uint32 flags=DEFAULT)=0;
		//! Set the alpha properties of the surface
		virtual void setAlpha(bool usePerPixelAlpha=false, Uint8 alphaValue=ALPHA_OPAQUE)=0;
		//! Return the width of the surface
		virtual int getW(void)=0;
		//! Return the height of the surface
		virtual int getH(void)=0;
		//! Return the depth of the surface
		virtual int getDepth(void)=0;
		//! Return the flags of the surface
		virtual int getFlags(void)=0;
		//! Set the clipping rectangle
		virtual void setClipRect(int x, int y, int w, int h)=0;
		//! Clear the clipping rectangle (set it to surface dimension)
		virtual void setClipRect(void)=0;
		//! Load image name on the surface
		virtual void loadImage(const char *name)=0;
		//! Draw sprite frame index at (x,y)
		virtual void drawSprite(int x, int y, Sprite *sprite, int index=0)=0;
		//! Set pixel (x,y) to color (r,g,b,a)
		virtual void drawPixel(int x, int y, Uint8 r, Uint8 g, Uint8 b, Uint8 a=ALPHA_OPAQUE)=0;
		//! Draw an empty rectangle from (x,y) to (x+w,y+h) using color (r,g,b,a)
		virtual void drawRect(int x, int y, int w, int h, Uint8 r, Uint8 g, Uint8 b, Uint8 a=ALPHA_OPAQUE)=0;
		//! Draw a filled rectangle from (x,y) to (x+w,y+h) using color (r,g,b,a)
		virtual void drawFilledRect(int x, int y, int w, int h, Uint8 r, Uint8 g, Uint8 b, Uint8 a=ALPHA_OPAQUE)=0;
		//! Draw a vertical line from (x,y) to (x,y+l) using color (r,g,b,a)
		virtual void drawVertLine(int x, int y, int l, Uint8 r, Uint8 g, Uint8 b, Uint8 a=ALPHA_OPAQUE)=0;
		//! Draw a horizonzal line from (x,y) to (x+l,y) using color (r,g,b,a)
		virtual void drawHorzLine(int x, int y, int l, Uint8 r, Uint8 g, Uint8 b, Uint8 a=ALPHA_OPAQUE)=0;
		//! Draw a line from (x1,y1) to (x2,y2) using color (r,g,b,a)
		virtual void drawLine(int x1, int y1, int x2, int y2, Uint8 r, Uint8 g, Uint8 b, Uint8 a=ALPHA_OPAQUE)=0;
		//! Draw an empty circle of radius ray centered in (x,y) using color (r,g,b,a)
		virtual void drawCircle(int x, int y, int ray, Uint8 r, Uint8 g, Uint8 b, Uint8 a=ALPHA_OPAQUE)=0;
		//! Draw text of value i at (x,y) using font
		virtual void drawString(int x, int y, const Font *font, int i)=0;
		//! Draw msg at (x,y) using font
		virtual void drawString(int x, int y, const Font *font, const char *msg)=0;
		//! Draw msg from (x,y) with width w using font 
		virtual void drawString(int x, int y, int w, const Font *font, const char *msg)=0;
		//! Draw surface at (x,y)
		virtual void drawSurface(int x, int y, DrawableSurface *surface)=0;
		//! Blit to screen rectangles rects, which is an array of length size
		virtual void updateRects(SDL_Rect *rects, int size)=0;
		//! Blit to screen rectangle from (x,y) to (x+w,y+h)
		virtual void updateRect(int x, int y, int w, int h)=0;
		//! Return a pointer to pixels of the underlying surface
		virtual void *getPixelPointer(void) = 0;
	};
	
	//! A DrawableSurface which is the window of the application
	//! \ingroup libgag
	class GraphicContext:public virtual DrawableSurface
	{
	private:
		SDL_Rect **modes; //!< available graphic modes (such as 640x480, 800x600, ...)
		
	protected:
		int minW; //!< minimum width of the window/screen
		int minH; //!< minimum height of the window/screen
	
	public:
		//! Constructor, do not open the window
		GraphicContext();
		//! Destructor, close the window if opened (if setRes() has been called)
		virtual ~GraphicContext(void);
	
		//! Create the window. This must be called before any Drawable Surface method.
		virtual bool setRes(int w, int h, int depth=32, Uint32 flags=DEFAULT)=0;
		//! Defined the minimum acceptable resolution for the window
		virtual void setMinRes(int w=0, int h=0);
		//! Initialise the listing of video modes
		virtual void beginVideoModeListing(void);
		//! Put the next available video mode in w and h, return true if there is still more available video modes
		virtual bool getNextVideoMode(int *w, int *h);
		//! Set the title and status bar caption of the window
		virtual void setCaption(const char *title, const char *icon)=0;
	
		//! Load sprite from filename
		virtual Sprite *loadSprite(const char *filename);
	
		//! Load font of size from filename
		virtual Font *loadFont(const char *filename, unsigned size)=0;
	
		//! Create a new off-screen drawable surface. Load image name on it
		virtual DrawableSurface *createDrawableSurface(const char *name=NULL)=0;
	
		//! Blit to screen
		virtual void nextFrame(void)=0;
		//! Print screen to filename
		virtual void printScreen(const char *filename)=0;
	
		//! Create a graphic context
		static GraphicContext *createGraphicContext();
	};
	
	//! Font with a given foundery, shape and color
	//! \ingroup libgag
	class Font
	{
	public:
		//! Shape of the font
		enum Shape
		{
			STYLE_NORMAL = 0x00, //!< normal fond
			STYLE_BOLD = 0x01, //!< bold font
			STYLE_ITALIC = 0x02, //!< italic font
			STYLE_UNDERLINE = 0x04, //!< underlined font
		};
	
	public:
		//! Virtual destructor, no nothing
		virtual ~Font() { }
	
		// width and height
		//! Return width of string in pixels
		virtual int getStringWidth(const char *string) const=0;
		//! Return width of len first pixels of string in pixels
		virtual int getStringWidth(const char *string, int len) const;
		//! Return width of i in pixels
		virtual int getStringWidth(const int i) const;
		//! Return height of string in pixels
		virtual int getStringHeight(const char *string) const=0;
		//! Return height of len first pixels of string in pixels
		virtual int getStringHeight(const char *string, int len) const;
		//! Return width of i in pixels
		virtual int getStringHeight(const int i) const;
		//! Return true if c is a printable character
		virtual bool printable(char c) const=0;
	
		// Style and color
		//! Set the font color, clear the color stack
		virtual void setColor(Uint8 r, Uint8 g, Uint8 b, Uint8 a = DrawableSurface::ALPHA_OPAQUE) { }
		//! Push the font color on the color stack and set it
		virtual void pushColor(Uint8 r, Uint8 g, Uint8 b, Uint8 a = DrawableSurface::ALPHA_OPAQUE) { }
		//! Remove the last element of the color stack
		virtual void popColor(void) { }
		//! Return the actual font color
		virtual void getColor(Uint8 *r, Uint8 *g, Uint8 *b, Uint8 *a) const { }
	
		//! Set the font style, clear the style stack
		virtual void setStyle(unsigned style) { }
		//! Push the font style on the style stack and set it
		virtual void pushStyle(unsigned style) { }
		//! Remove the last element of the style stack
		virtual void popStyle(void) { }
		//! Return the actual font style
		virtual unsigned getStyle(void) const { return 0; }
	};
	
	//! Color is both 4 one byte components and a 4 bytes unsigned int
	//! \ingroup libgag
	union Color32
	{
		Uint32 id; //!< access as one 4 bytes unsigned int
		struct
		{
			Uint8 r, g, b, a; //!< components
		} channel; //!< access as 4 one byte components
	
		//! Constructor. Default color is opaque black
		Color32() { channel.r=channel.g=channel.b=0; channel.a=DrawableSurface::ALPHA_OPAQUE; }
		//! Constructor from components
		Color32(Uint8 r, Uint8 g, Uint8 b, Uint8 a=DrawableSurface::ALPHA_OPAQUE) { channel.r=r; channel.g=g; channel.b=b; channel.a=a; }
		//! Constructor from unsigned int
		Color32(Uint32 v) { id=v; }
		//! Return true if id < o.id
		bool operator<(const Color32 &o) const { return id<o.id; }
	};
	
	//! A sprite is a collection of images (frames) that can be displayed one after another to make an animation
	//! \ingroup libgag
	class Sprite
	{
	protected:
		//! Struct holding an SDL surface
		struct Surface
		{
			SDL_Surface *s; //!< SDL surface
	
			//! Constructor, allocate the internal surface suitable for fast blit, free the source
			Surface(SDL_Surface *source);
			//! Destructor, release SDL surface
			~Surface();
		};
	
		//! Struct holding a surface and its rotated versions around hue
		struct RotatedImage
		{
			SDL_Surface *orig; //!< Original surface
			typedef std::map<Color32, Surface *> RotationMap; //!< Map of colors to rotated surfaces
			RotationMap rotationMap; //!< Rotated surfaces
	
			//! Constructor, create rotated surfaces from s
			RotatedImage(SDL_Surface *s) { orig=s; }
			//! Destructor, release rotated surfaces
			~RotatedImage();
		};
	
		std::vector <Surface *> images; //!< Non-rotated frames
		std::vector <RotatedImage *> rotated; //!< Rotated frames
		Color32 actColor; //!< Starting color from which hue rotation will be defined
	
		friend class GraphicContext;
		//! Load both types of frames (rotated and not rotated ones) from streams
		void loadFrame(SDL_RWops *frameStream, SDL_RWops *rotatedStream);
		//! Create a surface from an SDL one
		Surface *surfaceFromSDL(SDL_Surface *s);
	
	public:
		//! Constructor, empty sprite
		Sprite() { }
		//! Destructor, release all frames
		virtual ~Sprite();
	
		//! Draw the sprite frame index at pos (x,y) on an SDL Surface with the clipping rect clip
		virtual void drawSDL(SDL_Surface *dest, const SDL_Rect *clip, int x, int y, int index);
		
		//! Set the (r,g,b) color to a sprite's base color
		virtual void setBaseColor(Uint8 r, Uint8 g, Uint8 b) { actColor=Color32(r, g, b); }
	
		//! Return the width of index frame of the sprite
		virtual int getW(int index);
		//! //! Return the height of index frame of the sprite
		virtual int getH(int index);
	};
}

#endif
