/*
  Copyright (C) 2001-2005 Stephane Magnenat & Luc-Olivier de Charrière
  for any question or comment contact us at nct@ysagoon.com or nuage@ysagoon.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "SDLGraphicContext.h"
#include <math.h>
#include <assert.h>
#include <SDL_image.h>

#ifdef HAVE_CONFIG_H
	#include <config.h>
#endif

/*!	\file Sprite.cpp
	\brief Implementation of sprite using SDL
*/

namespace GAG
{
	//! Global pointer to screen
	extern SDLGraphicContext *screen;
	
	//! Return the min of f1,f2 and f3
	float fmin(float f1, float f2, float f3)
	{
		if ((f1<=f2) && (f1<=f3))
			return f1;
		else if (f2<=f3)
			return f2;
		else
			return f3;
	}
	
	//! Return the max of f1,f2 and f3
	float fmax(float f1, float f2, float f3)
	{
		if ((f1>=f2) && (f1>=f3))
			return f1;
		else if (f2>=f3)
			return f2;
		else
			return f3;
	}
	
	//! Transform (r,g,b) in RGB to (h,s,v) in HSV (Hue,Saturation,Value)
	void RGBtoHSV( float r, float g, float b, float *h, float *s, float *v )
	{
		float min, max, delta;
		min = fmin( r, g, b );
		max = fmax( r, g, b );
		*v = max;				// v
		delta = max - min;
		if( max != 0 )
			*s = delta / max;		// s
		else {
			// r = g = b = 0		// s = 0, v is undefined
			*s = 0;
			*h = -1;
			return;
		}
		if( r == max )
			*h = ( g - b ) / delta;		// between yellow & magenta
		else if( g == max )
			*h = 2 + ( b - r ) / delta;	// between cyan & yellow
		else
			*h = 4 + ( r - g ) / delta;	// between magenta & cyan
		*h *= 60;				// degrees
		if( *h < 0 )
			*h += 360;
	}
	
	//! Transform (h,s,v) in HSV (Hue,Saturation,Value) to (r,g,b) in RGB
	void HSVtoRGB( float *r, float *g, float *b, float h, float s, float v )
	{
		int i;
		float f, p, q, t;
		if( s == 0 ) {
			// achromatic (grey)
			*r = *g = *b = v;
			return;
		}
		h /= 60;			// sector 0 to 5
		i = (int)floor( h );
		f = h - i;			// factorial part of h
		p = v * ( 1 - s );
		q = v * ( 1 - s * f );
		t = v * ( 1 - s * ( 1 - f ) );
		switch( i ) {
			case 0:
				*r = v;
				*g = t;
				*b = p;
				break;
			case 1:
				*r = q;
				*g = v;
				*b = p;
				break;
			case 2:
				*r = p;
				*g = v;
				*b = t;
				break;
			case 3:
				*r = p;
				*g = q;
				*b = v;
				break;
			case 4:
				*r = t;
				*g = p;
				*b = v;
				break;
			default:		// case 5:
				*r = v;
				*g = p;
				*b = q;
				break;
		}
	}
	
	Sprite::Surface::Surface(SDL_Surface *source)
	{
		assert(source);
		s = SDL_DisplayFormatAlpha(source);
		assert(s);
		SDL_FreeSurface(source);
	
	}
	
	Sprite::Surface::~Surface()
	{
		SDL_FreeSurface(s);
	}
	
	Sprite::RotatedImage::~RotatedImage()
	{
		SDL_FreeSurface(orig);
		for (RotationMap::iterator it = rotationMap.begin(); it != rotationMap.end(); ++it)
		{
			delete it->second;
		}
	}
	
	
	void Sprite::drawSDL(SDL_Surface *dest, const SDL_Rect *clip, int x, int y, int index)
	{
		if ((index<0) || (index>=(int)images.size()))
			fprintf(stderr, "GAG : Can load index %d of %u\n", index, images.size());
		assert(index>=0);
		assert(index<(int)images.size());
	
		SDL_Rect oldr, r;
		SDL_Rect newr=*clip;
		SDL_Rect src;
		int w, h;
		int diff;
	
		w=getW(index);
		h=getH(index);
	
		src.x=0;
		src.y=0;
		if (x<newr.x)
		{
			diff=newr.x-x;
			w-=diff;
			src.x+=diff;
			x=newr.x;
		}
		if (y<newr.y)
		{
			diff=newr.y-y;
			h-=diff;
			src.y+=diff;
			y=newr.y;
		}
	
		if (x+w>newr.x+newr.w)
		{
			diff=(x+w)-(newr.x+newr.w);
			w-=diff;
		}
		if (y+h>newr.y+newr.h)
		{
			diff=(y+h)-(newr.y+newr.h);
			h-=diff;
		}
	
		if ((w<=0) || (h<=0))
			return;
	
		src.w=w;
		src.h=h;
		r.x=x;
		r.y=y;
		r.w=w;
		r.h=h;
	
		SDL_GetClipRect(dest, &oldr);
		SDL_SetClipRect(dest, &newr);
	
		if (images[index])
			SDL_BlitSurface(images[index]->s, &src, dest, &r);
	
		if (rotated[index])
		{
			RotatedImage::RotationMap::const_iterator it = rotated[index]->rotationMap.find(actColor);
			Surface *toBlit;
			if (it == rotated[index]->rotationMap.end())
			{
				float hue, lum, sat;
				float baseHue, hueDec;
	
				RGBtoHSV(51.0f/255.0f, 255.0f/255.0f, 153.0f/255.0f, &baseHue, &sat, &lum);
				RGBtoHSV( ((float)actColor.channel.r)/255, ((float)actColor.channel.g)/255, ((float)actColor.channel.b)/255, &hue, &sat, &lum);
				hueDec = hue-baseHue;
				int w = rotated[index]->orig->w;
				int h = rotated[index]->orig->h;
	
				SDL_Surface *newSurface = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, 32, 0xff, 0xff00, 0xff0000, 0xff000000);
	
				Uint32 *sPtr = (Uint32 *)rotated[index]->orig->pixels;
				Uint32 *dPtr = (Uint32 *)newSurface->pixels;
				for (int i=0; i<w*h; i++)
				{
					Uint8 sR, sG, sB, alpha;
					Uint8 dR, dG, dB;
					float nR, nG, nB;
					SDL_GetRGBA(*sPtr, rotated[index]->orig->format, &sR, &sG, &sB, &alpha);
	
					if (alpha != DrawableSurface::ALPHA_TRANSPARENT)
					{
						RGBtoHSV( ((float)sR)/255, ((float)sG)/255, ((float)sB)/255, &hue, &sat, &lum);
	
						float newHue = hue + hueDec;
						if (newHue >= 360)
							newHue -= 360;
						if (newHue < 0)
							newHue += 360;
	
						HSVtoRGB(&nR, &nG, &nB, newHue, sat, lum);
	
						dR = (Uint32)(255*nR);
						dG = (Uint32)(255*nG);
						dB = (Uint32)(255*nB);
	
						*dPtr = SDL_MapRGBA(newSurface->format, dR, dG, dB, alpha);
					}
					else
						*dPtr = 0;
	
	
					sPtr++;
					dPtr++;
				}
	
				toBlit = new Surface(newSurface);
				rotated[index]->rotationMap[actColor] = toBlit;
			}
			else
			{
				toBlit = it->second;
			}
			SDL_BlitSurface(toBlit->s, &src, dest, &r);
		}
	
		SDL_SetClipRect(dest, &oldr);
	}
	
	Sprite::~Sprite()
	{
		for (std::vector <Surface *>::iterator imagesIt=images.begin(); imagesIt!=images.end(); ++imagesIt)
		{
			if (*imagesIt)
				delete (*imagesIt);
		}
		for (std::vector <RotatedImage *>::iterator rotatedIt=rotated.begin(); rotatedIt!=rotated.end(); ++rotatedIt)
		{
			if (*rotatedIt)
				delete (*rotatedIt);
		}
	}
	
	void Sprite::loadFrame(SDL_RWops *frameStream, SDL_RWops *rotatedStream)
	{
		if (frameStream)
		{
			SDL_Surface *temp = IMG_Load_RW(frameStream, 0);
			images.push_back(new Surface(temp));
		}
		else
			images.push_back(NULL);
	
		if (rotatedStream)
		{
			SDL_Surface *sprite = IMG_Load_RW(rotatedStream, 0);
			assert(sprite);
			if (sprite->format->BitsPerPixel==32)
			{
				RotatedImage *image = new RotatedImage(sprite);
				rotated.push_back(image);
			}
			else
			{
				fprintf(stderr, "GAG : Warning, rotated image is in wrong for (%d) bpp istead of 32\n", sprite->format->BitsPerPixel);
				rotated.push_back(NULL);
			}
		}
		else
			rotated.push_back(NULL);
	}
	
	int Sprite::getW(int index)
	{
		assert(index>=0);
		assert(index<(int)images.size());
		if (images[index])
			return images[index]->s->w;
		else if (rotated[index])
			return rotated[index]->orig->w;
		else
			return 0;
	}
	
	int Sprite::getH(int index)
	{
		assert(index>=0);
		assert(index<(int)images.size());
		if (images[index])
			return images[index]->s->h;
		else if (rotated[index])
			return rotated[index]->orig->h;
		else
			return 0;
	}
}
