#include "Registry.h"
#include "../ga/GenerationalEvolution.h"
#include "../ga/SteadyStateEvolution.h"
#include "../experiment/Experiment.h"
#include "../experiment/NullExperiment.h"
#include "../experiment/KheperaNavigation.h"
#include "../experiment/BatteryRecharge.h"
#include "../controller/nn/SigmoidNeuron.h"
#include "../controller/nn/SRMNeuron.h"
#include "../controller/GenotypeDecoder.h"
#include "../controller/ParameterGenotypeDecoder.h"
#include "../controller/nn/GraphNeuralNetwork.h"
#include "../controller/MatrixController.h"
#include "../controller/nn/FeedForwardNeuralNetwork.h"
#include <string>

#ifdef OPT_MODULE_HEADER_INCLUDE
#include OPT_MODULE_HEADER_INCLUDE
#endif

/*!	\file Registry.cpp
	\brief Creation, populing and destruction of all libraries
*/

namespace Teem
{
	#ifdef OPT_MODULE_PROTO_INLINE
	OPT_MODULE_PROTO_INLINE
	#endif

	//! Library of all available evolutions
	Registrar<EvolutionRun, std::string> *evolutionsLibrary = NULL;
	//! Library of all available experiment
	Registrar<Experiment, std::string> *experimentsLibrary = NULL;
	//! Library of all available neurons
	Registrar<Neuron, std::string> *neuronsLibrary = NULL;
	//! Library of all available synapses
	Registrar<Synapse, std::string> *synapsesLibrary = NULL;
	//! Library of all available genotype decoders
	Registrar<GenotypeDecoder, std::string, const std::string&> *genotypeDecodersLibrary = NULL;

	void initRegistry(void)
	{
		// create the registrars
		evolutionsLibrary = new Registrar<EvolutionRun, std::string>;
		experimentsLibrary = new Registrar<Experiment, std::string>;
		neuronsLibrary = new Registrar<Neuron, std::string>;
		synapsesLibrary = new Registrar<Synapse, std::string>;
		genotypeDecodersLibrary = new Registrar<GenotypeDecoder, std::string, const std::string&>;

		// register experiments
		#ifdef OPT_MODULE_CODE_INLINE
		OPT_MODULE_CODE_INLINE
		#endif
		#ifdef OPT_MODULE_CODE_INCLUDE
		#include OPT_MODULE_CODE_INCLUDE
		#endif

		// register evolution runs
		evolutionsLibrary->registerItem<SimpleEvolutionRun>("Simple");
		evolutionsLibrary->registerItem<SimpleMultiGenomeEvolutionRun>("SimpleMultiGenome");
		evolutionsLibrary->registerItem<SteadyStateEvolutionRun>("SteadyState");

		// register neurons
		neuronsLibrary->registerItem<CurrentInputNeuron>("CurrentInput");
		neuronsLibrary->registerItem<SimpleSpikeInputNeuron>("SimpleSpikeInput");
		neuronsLibrary->registerItem<SigmoidNeuron>("Sigmoid");
		neuronsLibrary->registerItem<SigmoidInputNeuron>("SigmoidInput");
		neuronsLibrary->registerItem<FiringRateInputNeuron>("FiringRateInput");
		neuronsLibrary->registerItem<FiringRateOutputNeuron>("FiringRateOutput");
		neuronsLibrary->registerItem<SRMNeuron>("SRM");

		// register synapses
		synapsesLibrary->registerItem<Synapse>("Simple");
		synapsesLibrary->registerItem<SRMSynapse>("SRM");
		synapsesLibrary->registerItem<SRMPlasticSynapse>("SRMPlastic");

		// register genotype decoders
		genotypeDecodersLibrary->registerItem<ParameterGenotypeDecoder>("Parameter");
		genotypeDecodersLibrary->registerItem<GraphFeedForwardGenotypeDecoder>("GraphFeedForward");
		genotypeDecodersLibrary->registerItem<GraphFullyConnectedGenotypeDecoder>("GraphFullyConnected");
		genotypeDecodersLibrary->registerItem<MatrixGenotypeDecoder>("Matrix");
		genotypeDecodersLibrary->registerItem<FeedForwardNeuralNetworkGenotypeDecoder>("FeedForwardNeuralNetwork");
		genotypeDecodersLibrary->registerItem<DummyGenotypeDecoder>("Dummy");
	}

	void destroyRegistry(void)
	{
		// destroy registrars
		delete evolutionsLibrary;
		delete experimentsLibrary;
		delete neuronsLibrary;
		delete synapsesLibrary;
		delete genotypeDecodersLibrary;
	}
}
