#include <fstream>
#include <iomanip>

#include "Graph.h"

/*!	\file Graph.cpp
	\brief Implementation of classes to create graphs
*/

namespace Teem
{
	IshtarSettingsMap::IshtarSettingsMap(const std::string &root)
	{
		typedef std::map<std::string, Ishtar::ParsedValues> ParsedValuesType;
		
		
		// BRUTE FORCE IMPLEMENTATION --- NOT OPTIMAL
		
		for(ParsedValuesType::iterator i = Ishtar::settings->parsedTable.begin();
			i != Ishtar::settings->parsedTable.end(); i++)
		{
			if(i->first.find(root + ".") == 0)
			{
				std::string s = i->first.substr(root.size()+1);
				assert(i->second.size() == 1);
				std::istringstream iss(i->second[0]);
				double d;
				iss >> d;
				(*this)[s] = d;
			}
		}
		
		
		// GOOD BUT NON WORKING IMPLEMENTATION...
		// TODO: fix and use it..
		
	/*
		// use an empty pair... the value is passed to the comparison functor
		std::pair<std::string, Ishtar::ParsedValues> p;
		Comparator comp(root + ".");
		
		ParsedValuesType::iterator begin = std::lower_bound(
									Ishtar::settings->parsedTable.begin(),
									Ishtar::settings->parsedTable.end(),
									p, comp);
		ParsedValuesType::iterator end = std::upper_bound(
									begin,
									Ishtar::settings->parsedTable.end(),
									p, comp);
		
		for(ParsedValuesType::iterator i = begin; i != end; i++)
		{
			std::string s = i->first;
			s = s.substr(root.size() + 1);
			assert(i->second.size() == 1);
			std::istringstream iss(i->second[0]);
			double d;
			iss >> d;
			(*this)[s] = d;
		}
	*/
	}
	
	Graph::Node Graph::createNode(const std::string &type, const std::set<std::string> &tags)
	{
		// Setup data structure.
		NodeData data;
		data.type = type;
		data.tags = tags;
		data.params.insert(defaultNodeParams.begin(), defaultNodeParams.end());
		
		// Add to vector.
		nodes.push_back(data);
		Node node = nodes.size() - 1;
		
		// Maintain index lists.
		for(std::set<std::string>::iterator i = tags.begin(); i != tags.end(); i++)
			nodeLists[*i].push_back(node);
		
		return node;
	}
	
	Graph::Link Graph::createLink(Node from, Node to, const std::string &type, const std::set<std::string> &tags)
	{
		// Setup data structure.
		LinkData data;
		data.from = from;
		data.to = to;
		data.type = type;
		data.tags = tags;
		data.params.insert(defaultLinkParams.begin(), defaultLinkParams.end());
		
		// Add to vector.
		links.push_back(data);
		Link link = links.size() - 1;
		
		// Maintain index lists.
		for(std::set<std::string>::iterator i = tags.begin(); i != tags.end(); i++)
			linkLists[*i].push_back(link);
		
		return link;
	}
	
	void Graph::exportToDot(const std::string &filename) const
	{
		std::ofstream out(filename.c_str());
		
		if (out.good())
		{
			out << "digraph GraphNeuralNetwork\n{\n";
			out << "\tsize=\"6,6\"; ratio = fill;\n";
			for (size_t i=0; i<nodes.size(); i++)
			{
				std::ostringstream neuronDotArg;
				neuronDotArg << "\"" << i << " (" << nodes[i].type << ")\"";
				out << "\tneuron" << i << " [shape=box,label=" << neuronDotArg.str() << "];\n";
			}
			for (size_t i=0; i<links.size(); i++)
			{
				std::map<std::string, double>::const_iterator weightIt = links[i].params.find(std::string("weight"));
				if (weightIt != links[i].params.end())
				{
					double weight = weightIt->second;
					if (weight > 0)
					{
						if (weight > 0.5)
							out << "\tneuron" << links[i].from << " -> neuron" << links[i].to << std::setprecision(3) << " [color=black,style=bold,label=\"" << weight << "\"];\n";
						else
							out << "\tneuron" << links[i].from << " -> neuron" << links[i].to << std::setprecision(3) << " [color=black,style=solid,label=\"" << weight << "\"];\n";
					}
					else
					{
						if (weight < -0.5)
							out << "\tneuron" << links[i].from << " -> neuron" << links[i].to << std::setprecision(3) << " [color=lightgrey,style=bold,label=\"" << weight << "\"];\n";
						else
							out << "\tneuron" << links[i].from << " -> neuron" << links[i].to << std::setprecision(3) << " [color=lightgrey,style=solid,label=\"" << weight << "\"];\n";
					}
				}
				else
					out << "\tneuron" << links[i].from << " -> neuron" << links[i].to << " [color=black,style=solid];\n";
			}
			out << "}\n";
			out.close();
		}
		else
		{
			std::cerr << "Graph::exportToDot(" << filename << ") : can't open file" << std::endl;
		}
	}
	
	void Graph::exportToMatlab(const std::string &filename) const
	{
		std::ofstream out(filename.c_str());
		
		if (out.good())
		{
			out << "numNeuron = " << nodes.size() << ";\n";
			out << "numSynapse = " << nodes.size() << ";\n";
			
			// save neuron lists
			for(std::map<std::string, std::vector<Node> >::const_iterator i = nodeLists.begin(); i != nodeLists.end(); i++)
			{
				out << i->first << "Neurons = [";
				for(size_t j = 0; j < i->second.size(); j++)
					out << i->second[j] + 1 << " ";
				out << "];\n";
			}
			
			// save synapses.
			out << "synapses = [";
			for(size_t i = 0; i < links.size(); i++)
			{
				out << links[i].from + 1 << " " << links[i].to + 1 << "; ";
			}
			out << "];\n";
			
			// save synapse lists
			for(std::map<std::string, std::vector<Link> >::const_iterator i = linkLists.begin(); i != linkLists.end(); i++)
			{
				out << i->first << "Synapses = [";
				for(size_t j = 0; j < i->second.size(); j++)
					out << i->second[j] + 1 << " ";
				out << "];\n";
			}
			out.close();
		}
		else
		{
			std::cerr << "Graph::exportToMatlab(" << filename << ") : can't open file" << std::endl;
		}
	}
}
