/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifdef HAVE_ENKI

#include "Enki.h"
#include "../utils/VFS.h"

#include <enki/PhysicalEngine.h>
#include <enki/robots/s-bot/Sbot.h>
#include <enki/robots/s-bot/SbotObject.h>
#include <enki/robots/alice/Alice.h>
#include <enki/robots/khepera/Khepera.h>
#include <enki/robots/mav/Mav.h>
#include <assert.h>
#include <iostream>
#include <fstream>

#ifdef HAVE_VIEWER
// This viewer is a hack... beeuhhh
#include "../enkiviewer/EnkiViewer.h"
#endif

/*!	\file Enki.cpp
	\brief Implementation of world interface to Enki
*/

namespace Teem
{
	using namespace An;

	//! Load textures from file whose name is stored fileNameIshtarVariable. textures will be assigned from texturesBegin to texturesEnd
	void loadTexturesFromFile(An::Texture *texturesBegin, An::Texture *texturesEnd, const std::string &fileNameIshtarVariable)
	{
		Ishtar::Variable<std::string> bareFileName(fileNameIshtarVariable, "");
		if (((const std::string &)bareFileName) != "")
		{
			std::string fileName = vfs.getFullPath(bareFileName);
			std::ifstream ifs(fileName.c_str());
			if (ifs.is_open())
			{
				unsigned texCount = 0;
				for (An::Texture *iterator = texturesBegin; iterator != texturesEnd; iterator++)
				{
					unsigned len;
					ifs >> len;
					if (len != 0)
					{
						iterator->resize(len);
						for (unsigned i=0; i<len; i++)
							for (unsigned j=0; j<3; j++)
								ifs >> (*iterator)[i].components[j];
					}
					else
					{
						std::cerr << "Texture " << texCount << " of textures file " << fileName << " has zero length, filling with black." << std::endl;
						iterator->resize(1);
					}
					texCount++;
				}
			}
			else
				std::cerr << "Can't open textures file " << fileName << std::endl;
		}
	}

	//! Publish an instance of An::Vector through Ishtar
	inline void publishVector(An::Vector* vec, Ishtar::VariablesPublisher &publisher, const std::string &subRoot)
	{
		publisher.add<double, false, 2>(&vec->x, subRoot, Ishtar::StringsAccumulator() + "x" + "y");
	}

	//! Publish an instance of An::Color through Ishtar
	inline void publishColor(An::Color* color, Ishtar::VariablesPublisher &publisher, const std::string &subRoot)
	{
		publisher.add<double, false, 4>(&color->components[0], subRoot, Ishtar::StringsAccumulator() + "r" + "g" + "b" + "a");
	}

	//! Publish an instance Enki::PhysicalObject through Ishtar
	void publishPhysicalObject(Enki::PhysicalObject* op, Ishtar::VariablesPublisher &publisher)
	{
		publishVector(&op->pos, publisher, ".pos");
		publisher.add(&op->height, ".height");
		publisher.add(&op->angle, ".angle");
		publishVector(&op->speed, publisher, ".speed");
		publisher.add(&op->angSpeed, ".angSpeed");
		publisher.add(&op->mass, ".mass");
		publisher.add(&op->r, ".radius");
		publishColor(&op->color, publisher, ".color");
		publisher.add(&op->angle, ".angle");

	}

	// Interface to world in Enki
	EnkiWorld::EnkiWorld(bool createEnkiViewer) :
		ishtarPublisher("teem.world"),
		dimension("teem.world.dimension", 60),
		useVideo("teem.experiment.useVideo", 0),
		objectsArrayHelper("teem.world.objects"),
		realisticMode("teem.world.realistic", 0)
	{
		world = new Enki::World(dimension[0], dimension[1]);
		world->setRandomSeed(rand());
		viewer = NULL;

		ishtarPublisher.add(&world->useWalls, ".useWalls");

		#ifdef HAVE_VIEWER
		if(createEnkiViewer)
			createViewer<EnkiViewer>();
		#endif

		// create a registry of available objects
		Registrar<Entity, std::string, EnkiWorld *, std::string> entitiesLibrary;
		entitiesLibrary.registerItem<EnkiKhepera>("khepera");
		entitiesLibrary.registerItem<EnkiAlice>("alice");
		entitiesLibrary.registerItem<EnkiAliceSmallDisk>("aliceSmallDisk");
		entitiesLibrary.registerItem<EnkiAliceLargeDisk>("aliceLargeDisk");
		entitiesLibrary.registerItem<EnkiSbot>("sbot");
		entitiesLibrary.registerItem<EnkiSbotActiveObject>("sbotActiveObject");
		entitiesLibrary.registerItem<EnkiRectangularBlock>("rectangularBlock");

		// create objects from file
		for (unsigned i = 0; i<objectsArrayHelper.getMaxIndex(); i++)
		{
			std::string root = objectsArrayHelper.getRoot(i);
			Ishtar::Variable<std::string> objectType(root + ".type", "unknown");

			Entity *entity = entitiesLibrary.createItem((const std::string &)objectType, this, root);
			if (entity)
				autoCreatedEntities.push_back(entity);
			else
				std::cerr << "EnkiWorld::EnkiWorld : can't create object " << objectType << std::endl;
		}

		// load world textures
		loadTexturesFromFile(&world->wallTextures[0], &world->wallTextures[4], "teem.world.texturesFileName");
	}


	EnkiRealisticConditions::EnkiRealisticConditions():
		motorInstantNoiseMean("teem.world.motorInstantNoiseMean",  2.0188E-3),
		motorInstantNoiseStd("teem.world.motorInstantNoiseStd", 0.4532310),
		motorCoeffRange("teem.world.motorCoeffRange", 1.5/44.0),
		camNoiseMean("teem.world.camNoiseMean", -9.0E-3),//27.4277E-3),
		camNoiseStd("teem.world.camNoiseStd", 62.E-3), // exp= 32.8237E-3
		camProbBadPixel("teem.world.camProbaBadPixel", 1.0/128.0),
		motorCoeffGenerator(-motorCoeffRange, motorCoeffRange)
	{
	}

	double EnkiRealisticConditions::motorInstantNoise()
	{
		// In case the Ishtar values change over time, don't use a gaussian generator
		return An::gaussianRand(motorInstantNoiseMean, motorInstantNoiseStd);
	}

	double EnkiRealisticConditions::motorCoeff()
	{
		return (1.0+motorCoeffGenerator());
	}

	double EnkiRealisticConditions::camNoise()
	{
		return An::gaussianRand(camNoiseMean, camNoiseStd);
	}

	bool EnkiRealisticConditions::camBadPixel()
	{
		return An::boolRand(camProbBadPixel);
	}

	EnkiRealisticConditions::~EnkiRealisticConditions()
	{
	}

	//! S-bot realistic depth test, see through black
	struct SbotRealisticDepthTest : public Enki::PixelOperationFunctor
	{
		//! If objectDist2 < zBuffer2 && objectColor != black, then pixelBuffer = objectColor and zBuffer2 = objectDist2
		virtual void operator()(double &zBuffer2, An::Color &pixelBuffer, const double &objectDist2, const An::Color &objectColor)
		{
			if ((objectDist2 < zBuffer2) && (objectColor != An::Color::black))
			{
				zBuffer2 = objectDist2;
				pixelBuffer = objectColor;
			}
		}
	} sbotRealisticDepthTest; //!< Standard of S-bot realistic depth test, see through black


	EnkiWorld::~EnkiWorld()
	{
		#ifdef HAVE_VIEWER
		viewer->stopRecordingVideo();
		if(viewer)
			delete viewer;
		#endif

		// delete auto created entities
		for (size_t i=0; i<autoCreatedEntities.size(); i++)
			delete autoCreatedEntities[i];

		delete world;
	}


	Sbot *EnkiWorld::createSbot()
	{
		return new EnkiSbot(this, objectsArrayHelper.getNewRoot());
	}

	SbotActiveObject *EnkiWorld::createSbotActiveObject()
	{
		return new EnkiSbotActiveObject(this, objectsArrayHelper.getNewRoot());
	}

	Khepera *EnkiWorld::createKhepera()
	{
		return new EnkiKhepera(this, objectsArrayHelper.getNewRoot());
	}

	Alice *EnkiWorld::createAlice()
	{
		return new EnkiAlice(this, objectsArrayHelper.getNewRoot());
	}

	AliceSmallDisk *EnkiWorld::createAliceSmallDisk()
	{
		return new EnkiAliceSmallDisk(this, objectsArrayHelper.getNewRoot());
	}

	AliceLargeDisk *EnkiWorld::createAliceLargeDisk()
	{
		return new EnkiAliceLargeDisk(this, objectsArrayHelper.getNewRoot());
	}
        Mav *EnkiWorld::createMav()
        {
                return new EnkiMav(world, objectsArrayHelper.getNewRoot());
        }

	RectangularBlock *EnkiWorld::createRectangularBlock()
	{
		return new EnkiRectangularBlock(this, objectsArrayHelper.getNewRoot());
	}

	void EnkiWorld::step(double dt)
	{
		world->step(dt);

		#ifdef HAVE_VIEWER
		viewer->step(dt);
		#endif
	}

	An::Point EnkiWorld::getDimension(void)
	{
		return An::Point(world->w, world->h);
	}

	void EnkiWorld::setWallTextures(const std::valarray<An::Texture> &textures)
	{
		assert(textures.size() == 4);
		for (size_t i = 0; i < 4; i++)
		{
			world->wallTextures[i].resize(textures[i].size());
			world->wallTextures[i] = textures[i];
		}
	}

	void EnkiWorld::useWalls(bool use)
	{
		world->useWalls = use;
	}

	// Interface to Khepera in Enki
	EnkiKhepera::EnkiKhepera(EnkiWorld *world, const std::string &root) :
		ishtarPublisher(root)
	{
		object = new Enki::Khepera(Enki::Khepera::CAPABILITIY_BASIC_SENSORS | Enki::Khepera::CAPABILITY_CAMERA);
		assert(object);
		world->world->addObject(object);
		this->world = world;

		// Setup variable to be published via ishtar
		publishPhysicalObject(object, ishtarPublisher);
		ishtarPublisher.add(&object->leftSpeed, ".leftSpeed");
		ishtarPublisher.add(&object->rightSpeed, ".rightSpeed");
		ishtarPublisher.add(&object->leftEncoder, ".leftEncoder");
		ishtarPublisher.add(&object->rightEncoder, ".rightEncoder");
		ishtarPublisher.add(&object->infraredSensor0.finalValue, ".ir0");
		ishtarPublisher.add(&object->infraredSensor1.finalValue, ".ir1");
		ishtarPublisher.add(&object->infraredSensor2.finalValue, ".ir2");
		ishtarPublisher.add(&object->infraredSensor3.finalValue, ".ir3");
		ishtarPublisher.add(&object->infraredSensor4.finalValue, ".ir4");
		ishtarPublisher.add(&object->infraredSensor5.finalValue, ".ir5");
		ishtarPublisher.add(&object->infraredSensor6.finalValue, ".ir6");
		ishtarPublisher.add(&object->infraredSensor7.finalValue, ".ir7");
		ishtarPublisher.add<double, true, 50*4>(&(object->camera.image[0].components[0]), ".camera");
	}

	EnkiKhepera::~EnkiKhepera()
	{
		world->world->removeObject(object);
		delete object;
	}

	void EnkiKhepera::setSpeed(double left, double right)
	{
		object->leftSpeed = left;
		object->rightSpeed = right;
	}

	void EnkiKhepera::getEncoders(double &left, double &right)
	{
		left = object->leftEncoder;
		right = object->rightEncoder;
	}

	double EnkiKhepera::getIRSensor(unsigned index)
	{
		switch (index)
		{
			case 0: return object->infraredSensor0.finalValue;
			case 1: return object->infraredSensor1.finalValue;
			case 2: return object->infraredSensor2.finalValue;
			case 3: return object->infraredSensor3.finalValue;
			case 4: return object->infraredSensor4.finalValue;
			case 5: return object->infraredSensor5.finalValue;
			case 6: return object->infraredSensor6.finalValue;
			case 7: return object->infraredSensor7.finalValue;
			default: assert(false); return 0;
		}
	}


	void EnkiKhepera::setPosition(const An::Point &pos)
	{
		EnkiPositionable<Enki::Khepera>::setPosition(pos);
		object->resetEncoders();
	}

	unsigned EnkiKhepera::getCameraSize()
	{
		return object->camera.image.size();
	}

	Color EnkiKhepera::getCameraPixel(unsigned index)
	{
		assert(index < object->camera.image.size());
		return object->camera.image[object->camera.image.size() - index - 1];
	}

	// Interface to Alice in Enki

	EnkiAlice::EnkiAlice(EnkiWorld *world, const std::string &root) :
		ishtarPublisher(root)
	{
		object = new Enki::Alice(
			Enki::Alice::CAPABILITIY_BASIC_SENSORS |
			Enki::Alice::CAPABILITIY_EXT_HIGH_SENSOR |
			Enki::Alice::CAPABILITIY_CAMERA);
		assert(object);
		world->world->addObject(object);
		this->world = world;

		// setup variables to be published
		publishPhysicalObject(object, ishtarPublisher);
		ishtarPublisher.add(&object->leftSpeed, ".leftSpeed");
		ishtarPublisher.add(&object->leftSpeed, ".rightSpeed");
		ishtarPublisher.add(&object->left.finalValue, ".ir0");
		ishtarPublisher.add(&object->front.finalValue, ".ir1");
		ishtarPublisher.add(&object->right.finalValue, ".ir2");
		ishtarPublisher.add(&object->back.finalValue, ".ir3");
		ishtarPublisher.add(&object->frontHighExt.finalValue, ".ir4");
		ishtarPublisher.add(&object->backleft.finalValue, ".ir5");
		ishtarPublisher.add(&object->backright.finalValue, ".ir6");
		ishtarPublisher.add<double, true, 6*4>(&(object->circcam.image[0].components[0]), ".camera");
	}

	EnkiAlice::~EnkiAlice()
	{
		world->world->removeObject(object);
		delete object;
	}

	void EnkiAlice::setSpeed(double left, double right)
	{
		object->leftSpeed = left;
		object->rightSpeed = right;
	}

	void EnkiAlice::getSpeed(double &left, double &right)
	{
		left = object->speed.x;
		right = object->speed.y;
	}

	double EnkiAlice::getIRSensor(unsigned index)
	{
		switch (index)
		{
			case 0: return object->left.finalValue;
			case 1: return object->front.finalValue;
			case 2: return object->right.finalValue;
			case 3: return object->back.finalValue;
			case 4: return object->frontHighExt.finalValue;
			case 5: return object->backleft.finalValue;
			case 6: return object->backright.finalValue;
			default: assert(false); return 0;
		}
	}

	unsigned EnkiAlice::getCameraSize()
	{
		return object->circcam.image.size();
	}

	Color EnkiAlice::getCameraPixel(unsigned index)
	{
		assert(index < object->circcam.image.size());
		return object->circcam.image[index];
	}

	void EnkiAlice::setTransmitValue(unsigned val)
	{
		object->comm.setTransmitValue(val);
	}

	void EnkiAlice::setTransmit(bool enabled)
	{
		object->comm.setTransmit(enabled);
	}

	bool EnkiAlice::wasCommunication(void)
	{
		return object->comm.wasCommunication();
	}

	unsigned EnkiAlice::getSensorId(void)
	{
		return object->comm.getSensorId();
	}

	unsigned EnkiAlice::getSensorValue(void)
	{
		return object->comm.getSensorValue();
	}


	// Interface to AliceDisk in Enki
	EnkiAliceSmallDisk::EnkiAliceSmallDisk(EnkiWorld *world, const std::string &root) :
		ishtarPublisher(root)
	{
		object = new Enki::PhysicalObject();
		assert(object);

		object->r = 1.00; // old values: 1.05, .95
		object->mass = 10; // old value: 30
		
		object->staticFrictionThreshold = 0.01; // old values: 0.5
		object->height = 1.9;
		world->world->addObject(object);
		this->world = world;

		// setup variables to be published
		publishPhysicalObject(object, ishtarPublisher);
	}

	EnkiAliceSmallDisk::~EnkiAliceSmallDisk()
	{
		world->world->removeObject(object);
		delete object;
	}

	EnkiAliceLargeDisk::EnkiAliceLargeDisk(EnkiWorld *world, const std::string &root) :
		ishtarPublisher(root)
	{
		object = new Enki::PhysicalObject();
		assert(object);

		object->r = 3.5;
		object->mass = 40; // old value 50
		object->staticFrictionThreshold = 1.25; // old values 2.5, 10

		object->height = 1.9;
		world->world->addObject(object);
		this->world = world;

		// setup variables to be published
		publishPhysicalObject(object, ishtarPublisher);
	}

	EnkiAliceLargeDisk::~EnkiAliceLargeDisk()
	{
		world->world->removeObject(object);
		delete object;
	}

	// Interface to Sbot in Enki

	EnkiSbot::EnkiSbot(EnkiWorld *world, const std::string &root) :
		ishtarPublisher(root)
	{
		object = new Enki::FeedableSbot();
		world->world->addObject(object);
		this->world = world;
		if (world->realisticMode)
		{
			object->camera.setRange(150);
			object->camera.setPixelOperationFunctor(&sbotRealisticDepthTest);
		}

		// setup variables to be published
		publishPhysicalObject(object, ishtarPublisher);
		ishtarPublisher.add(&object->energy, ".energy");
		ishtarPublisher.add(&object->leftSpeed, ".leftSpeed");
		ishtarPublisher.add(&object->rightSpeed, ".rightSpeed");
		ishtarPublisher.add<double, true, 128*4>(&(object->camera.image[0].components[0]), ".camera");
	}

	EnkiSbot::~EnkiSbot()
	{
		world->world->removeObject(object);
		delete object;
	}

	void EnkiSbot::setSpeed(double left, double right)
	{
		if (world->realisticMode)
		{
			object->leftSpeed =
				( left
				+ world->realisticConditions.motorInstantNoise())
				* motorCoeffL;

			object->rightSpeed =
				( right
				+ world->realisticConditions.motorInstantNoise())
				* motorCoeffR;
		}
		else
		{
			object->leftSpeed  = left;
			object->rightSpeed = right;
		}
	}


	void EnkiSbot::getCameraAverage(An::Color meanPixels[8])
	{
		if (world->realisticMode)
		{
			static double sample[] = {1.0577, 1.16, 1.25, 1.225, 1.1667, 1.1667, 1.125, 1.0312, 0.96875, 1, 1, 0.89286, 1, 0.875, 0.875, 0.875, 0.875, 0.9, 0.9, 0.9, 0.85, 0.8, 0.75, 0.7, 0.75, 0.7, 0.75, 0.75, 0.75, 0.75, 0.75, 0.75, 0.625, 0.75, 0.6875, 0.625, 0.625, 0.625, 0.625, 0.625, 0.5625, 0.5625, 0.5625};

			for (size_t i=0; i<8; i++)
			{
				An::Color color; // color components are initialised to zero by default
				for (size_t j = (i*16); j<(i*16)+16; j++)
				{
					for (size_t k = 0; k < 3; k++)
					{
						// There is a probability that we see a totally wrong pixel
						if (world->realisticConditions.camBadPixel())
						{
							color.components[k] += rand()%2;
						}
						else
						{
							// Here we use t
							double c = object->camera.image[j].components[k];
							int distanceToObject = (int)round(sqrt(object->camera.zbuffer[j]));
							if (distanceToObject>100)
								c*=sample[(100-16)/2];
							else if (distanceToObject<16)
								c*=sample[0];
							else if (distanceToObject%2==0)
								c*=sample[(distanceToObject-16)/2];
							else
								c*=(sample[(distanceToObject-1-16)/2]
								    +sample[(distanceToObject+1-16)/2])/2.0;
							color.components[k] += c;
						}
					}
				}
				meanPixels[i] = color / 16;
 				meanPixels[i] += world->realisticConditions.camNoise();
			}
		}
		else
		{
			for (size_t i=0; i<8; i++)
			{
				An::Color color;
				for (size_t j = (i*16); j<(i*16)+16; j++)
					color += object->camera.image[j];
				meanPixels[i] = color / 16;
			}
		}
	}

	void EnkiSbot::getCamera(An::Color meanPixels[128])
	{
		std::copy(&object->camera.image[0], &object->camera.image[128], meanPixels);
	}

	void EnkiSbot::setRingColor(const An::Color &color)
	{
		object->color = color;
	}

	void EnkiSbot::clearEnergy(void)
	{
		object->energy = 0;
		object->dEnergy = 0;
		object->lastDEnergy = 0;
	}

	void EnkiSbot::setEnergy(double energy)
	{
		object->energy = energy;
	}

	double EnkiSbot::getEnergy(void)
	{
		return object->energy;
	}

	double EnkiSbot::getDEnergy(void)
	{
		return object->lastDEnergy;
	}

	void EnkiSbot::setFrequencies(unsigned frequencies)
	{
		object->sound.frequenciesState = frequencies;
	}

	unsigned EnkiSbot::getFrequencies(void)
	{
		return Enki::SbotGlobalSound::getWorldFrequenciesState();
	}

	void EnkiSbot::randomizeRealisticConditions(void)
	{
		if (world->realisticMode)
		{
			motorCoeffL = world->realisticConditions.motorCoeff();
			motorCoeffR = world->realisticConditions.motorCoeff();
		}
	}

	// Interface to Sbot Active Object in Enki

	EnkiSbotActiveObject::EnkiSbotActiveObject(EnkiWorld *world, const std::string &root) :
		ishtarPublisher(root)
	{
		object = new Enki::SbotActiveObject(10, 13);
		assert(object);
		//object = new Enki::SbotActiveObject(7, 15);
		world->world->addObject(object);
		this->world = world;

		// setup variables to be published
		publishPhysicalObject(object, ishtarPublisher);
		ishtarPublisher.add(&object->feeding.actualEnergy, ".energy");
		ishtarPublisher.add(&object->feeding.actualTime, ".actualTime");
		ishtarPublisher.add(&object->feeding.activeDuration, ".activeDuration");
		ishtarPublisher.add(&object->feeding.inactiveDuration, ".inactiveDuration");
		publishColor(&object->feeding.activeColor, ishtarPublisher, ".activeColor");
		publishColor(&object->feeding.activeColor, ishtarPublisher, ".inactiveColor");
		//ishtarPublisher.add<unsigned>((unsigned*) &object->feeding.consumeEnergy, ".consumeEnergy"); //FIXME check if bool correct
		ishtarPublisher.add(&object->feeding.consumeEnergy, ".consumeEnergy");
		ishtarPublisher.add(&object->feeding.dEnergyActive, ".dEnergyActive");
		ishtarPublisher.add(&object->feeding.dEnergyInactive, ".dEnergyInactive");
	}

	EnkiSbotActiveObject::~EnkiSbotActiveObject()
	{
		world->world->removeObject(object);
		delete object;
	}

	void EnkiSbotActiveObject::setEnergyProperties(double activeDuration, double inactiveDuration, bool consumeEnergy)
	{
		object->feeding.activeDuration = activeDuration;
		object->feeding.inactiveDuration = inactiveDuration;
		object->feeding.consumeEnergy = consumeEnergy;
	}

	void EnkiSbotActiveObject::setEnergyTime(double time)
	{
		object->feeding.actualTime = time;
	}

	void EnkiSbotActiveObject::setEnergyFlow(double activeFlow, double inactiveFlow)
	{
		object->feeding.dEnergyActive = activeFlow;
		object->feeding.dEnergyInactive = inactiveFlow;
	}

	void EnkiSbotActiveObject::setEnergy(double energy)
	{
		object->feeding.actualEnergy = energy;
	}

	double EnkiSbotActiveObject::getEnergy(void)
	{
		return object->feeding.actualEnergy;
	}

	void EnkiSbotActiveObject::setColor(const An::Color &activeColor, const An::Color &inactiveColor)
	{
		object->feeding.activeColor = activeColor;
		object->feeding.inactiveColor = inactiveColor;
	}

	double EnkiSbotActiveObject::getInteractionRange(void)
	{
		return object->feeding.getRange();
	}


	// Interface to a block in Enki

	EnkiRectangularBlock::EnkiRectangularBlock(EnkiWorld *world, const std::string &root) :
		ishtarPublisher(root)
	{
		object = new Enki::PhysicalObject();
		assert(object);
		world->world->addObject(object);
		this->world = world;

		// setup variables to be published
		publishPhysicalObject(object, ishtarPublisher);

		// default values
		set(10, 10, object->pos.x, object->pos.y, 0);
		object->height = 3;

		// publish bounding surface
		assert(boundingSurface.size() == 4);
		publishVector(&boundingSurface[0], ishtarPublisher, ".boundingSurfaceVertex.0");
		publishVector(&boundingSurface[1], ishtarPublisher, ".boundingSurfaceVertex.1");
		publishVector(&boundingSurface[2], ishtarPublisher, ".boundingSurfaceVertex.2");
		publishVector(&boundingSurface[3], ishtarPublisher, ".boundingSurfaceVertex.3");
		object->setBoundingSurface(&boundingSurface);

		// load texture
		loadTexturesFromFile(&object->textures[0], &object->textures[object->textures.size()], root + ".texturesFileName");
	}

	EnkiRectangularBlock::~EnkiRectangularBlock()
	{
		world->world->removeObject(object);
		delete object;
	}

	// Interface to Mav in Enki
	EnkiMav::EnkiMav(Enki::World *world, const std::string &root) :
		ishtarPublisher(root)
	{
		object = new Enki::Mav(
		Enki::Mav::CAPABILITIY_BASIC_SENSORS);
		world->addObject(object);
		this->world = world;

		// setup variables to be published
		publishPhysicalObject(object, ishtarPublisher);
		ishtarPublisher.add(&object->forwardSpeed, ".forwardSpeed");
		ishtarPublisher.add(&object->roll, ".roll");
	}

	EnkiMav::~EnkiMav()
	{
		world->removeObject(object);
		delete object;
	}

	void EnkiMav::setSpeed(double forwardSpeed, double roll)
	{
		object->forwardSpeed = forwardSpeed;
		object->roll = roll;
	}

	void EnkiMav::getSpeed(double &left, double &right)
	{
		left = object->speed.x;
		right = object->speed.y;
	}
	void EnkiMav::setTransmitValue( Enki::MavMessage *val)
	{
		object->comm.setTransmitValue(val);
	}
	
	Enki::MavMessage* EnkiMav::getTransmitValue(void)
	{
		return object->comm.getTransmitValue();
	}

	void EnkiMav::setTransmit(bool enabled)
	{
		object->comm.setTransmit(enabled);
	}

	bool EnkiMav::wasCommunication(void)
	{
		return object->comm.wasCommunication();
	}

	std::valarray<Enki::MavMessage *> EnkiMav::getSensorValue(void)
	{
		return object->comm.getSensorValue();
	}

	void EnkiRectangularBlock::set(double w, double h, double x, double y, double a)
	{
		object->pos = Point(x, y);
		object->angle = a;
		object->mass = -1;
		boundingSurface.clear();
		boundingSurface.push_back(Point(w*0.5, h*0.5));
		boundingSurface.push_back(Point(-w*0.5, h*0.5));
		boundingSurface.push_back(Point(-w*0.5, -h*0.5));
		boundingSurface.push_back(Point(w*0.5, -h*0.5));
		object->setBoundingSurface(&boundingSurface);
	}
}

#endif // HAVE_ENKI

